# -*- coding: utf-8 -*-
import logging
import utool as ut
import six
from wbia import _wbia_object
from wbia.control.controller_inject import make_ibs_register_decorator

(print, rrr, profile) = ut.inject2(__name__, '[images]')
logger = logging.getLogger('wbia')

CLASS_INJECT_KEY, register_ibs_method = make_ibs_register_decorator(__name__)

BASE_TYPE = type

try:
    from wbia import _autogen_image_base

    IMAGE_BASE = _autogen_image_base._image_base_class
except ImportError:
    IMAGE_BASE = _wbia_object.ObjectList1D

try:
    from wbia import _autogen_imageset_base

    IMAGESET_BASE = _autogen_imageset_base._imageset_base_class
except ImportError:
    IMAGESET_BASE = _wbia_object.ObjectList1D


@register_ibs_method
def images(ibs, gids=None, uuids=None, **kwargs):
    """ Makes an Images object """
    if uuids is not None:
        assert gids is None, 'specify one primary key'
        gids = ibs.get_image_gids_from_uuid(uuids)
    if gids is None:
        gids = ibs.get_valid_gids()
    elif gids.__class__.__name__ == 'Images':
        return gids
    gids = ut.ensure_iterable(gids)
    return Images(gids, ibs, **kwargs)


@register_ibs_method
def imagesets(ibs, gsids=None, text=None):
    if text is not None:
        gsids = ibs.get_imageset_imgsetids_from_text(text)
    if gsids is None:
        gsids = ibs.get_valid_imgsetids()
    elif gsids.__class__.__name__ == 'ImageSets':
        return gsids
    gsids = ut.ensure_iterable(gsids)
    return ImageSets(gsids, ibs)


class ImageIBEISPropertyInjector(BASE_TYPE):
    def __init__(metaself, name, bases, dct):
        super(ImageIBEISPropertyInjector, metaself).__init__(name, bases, dct)
        metaself.rrr = rrr
        # misc = [ 'instancelist', 'gids_with_aids', 'lazydict', ]  #
        attrs = [
            'aids',
            'aids_of_species',
            'annot_uuids',
            'annot_uuids_of_species',
            'annotation_bboxes',
            'annotation_thetas',
            'contributor_rowid',
            'contributor_tag',
            'datetime',
            'datetime_str',
            'detect_confidence',
            'detectpaths',
            'enabled',
            'exts',
            'gid',
            'glrids',
            'gnames',
            'gps',
            'gps2',
            'gsgrids',
            'heights',
            'imagesettext',
            'imgset_uuids',
            'imgsetids',
            'lat',
            'location_codes',
            'lon',
            'missing_uuid',
            'name_uuids',
            'nids',
            'notes',
            'num_annotations',
            'orientation',
            'orientation_str',
            'party_rowids',
            'party_tag',
            'paths',
            'reviewed',
            'sizes',
            'species_rowids',
            'species_uuids',
            'thumbpath',
            'thumbtup',
            'time_statstr',
            'timedelta_posix',
            'unixtime',
            'unixtime_asfloat',
            'unixtime2',
            'uris',
            'uris_original',
            'uuids',
            'widths' 'imgdata',
        ]
        # inverse_attrs = [
        #     'gids_from_uuid',
        # ]
        objname = 'image'
        _wbia_object._inject_getter_attrs(metaself, objname, attrs, [])


# @ut.reloadable_class
@six.add_metaclass(ImageIBEISPropertyInjector)
class Images(IMAGE_BASE):
    """
    Represents a group of annotations. Efficiently accesses properties from a
    database using lazy evaluation.

    CommandLine:
        python -m wbia.images Images --show

    Example:
        >>> # DISABLE_DOCTEST
        >>> from wbia.images import *  # NOQA
        >>> import wbia
        >>> ibs = wbia.opendb(defaultdb='testdb1')
        >>> gids = ibs.get_valid_gids()
        >>> g = self = images = Images(gids, ibs)
        >>> print(g.widths)
        >>> print(g)
        <Images(num=13)>
    """

    # def __init__(self, gids, ibs, config=None):
    #    super(Images, self).__init__(gids, ibs, config)

    @property
    def gids(self):
        return self._rowids

    @property
    def annots(self):
        return [self._ibs.annots(aids) for aids in self.aids]

    @property
    def _annot_groups(self):
        return self._ibs._annot_groups(self.annots)

    def remove_from_imageset(self, imageset_text):
        ibs = self._ibs
        if isinstance(imageset_text, six.string_types):
            gsid = ibs.get_imageset_imgsetids_from_text(imageset_text)
            gsids = [gsid] * len(self)
        else:
            gsids = ibs.get_imageset_imgsetids_from_text(imageset_text)
        ibs.unrelate_images_and_imagesets(self.gids, gsids)

    def append_to_imageset(self, imageset_text):
        ibs = self._ibs
        if isinstance(imageset_text, six.string_types):
            gsid = ibs.get_imageset_imgsetids_from_text(imageset_text)
            gsids = [gsid] * len(self)
        else:
            gsids = ibs.get_imageset_imgsetids_from_text(imageset_text)
        ibs.add_image_relationship(self.gids, gsids)

    def show(self, *args, **kwargs):
        if len(self) != 1:
            raise ValueError('Can only show one, got {}'.format(len(self)))
        from wbia.viz import viz_image

        for gid in self:
            return viz_image.show_image(self._ibs, gid, *args, **kwargs)


class ImageSetAttrInjector(BASE_TYPE):
    """
    Example:
        >>> # SCRIPT
        >>> from wbia import _wbia_object
        >>> import wbia
        >>> ibs = wbia.opendb(defaultdb='testdb1')
        >>> objname = 'imageset'
        >>> blacklist = []
        >>> _wbia_object._find_wbia_attrs(ibs, objname, blacklist)
    """

    def __init__(metaself, name, bases, dct):
        super(ImageSetAttrInjector, metaself).__init__(name, bases, dct)
        metaself.rrr = rrr
        # misc = [ 'instancelist', 'gids_with_aids', 'lazydict', ]  #
        attrs = [
            'aids',
            'configid',
            'custom_filtered_aids',
            'duration',
            'end_time_posix',
            'fraction_annotmatch_reviewed',
            'fraction_imgs_reviewed',
            'fraction_names_with_exemplar',
            'gids',
            'gps_lats',
            'gps_lons',
            'gsgrids',
            'image_uuids',
            'imgsetids_from_text',
            'imgsetids_from_uuid',
            'isoccurrence',
            'name_uuids',
            'nids',
            'note',
            'notes',
            'num_aids',
            'num_annotmatch_reviewed',
            'num_annots_reviewed',
            'num_gids',
            'num_imgs_reviewed',
            'num_names_with_exemplar',
            'percent_annotmatch_reviewed_str',
            'percent_imgs_reviewed_str',
            'percent_names_with_exemplar_str',
            'processed_flags',
            'shipped_flags',
            'smart_waypoint_ids',
            'smart_xml_contents',
            'smart_xml_fnames',
            'start_time_posix',
            'text',
            'uuid',
            'uuids',
        ]
        # inverse_attrs = [
        #     'gids_from_uuid',
        # ]
        objname = 'imageset'
        _wbia_object._inject_getter_attrs(metaself, objname, attrs, [])


# @ut.reloadable_class
@six.add_metaclass(ImageSetAttrInjector)
class ImageSets(IMAGESET_BASE):
    """
    Represents a group of annotations. Efficiently accesses properties from a
    database using lazy evaluation.

    CommandLine:
        python -m wbia.images ImageSets

    Example:
        >>> # DISABLE_DOCTEST
        >>> from wbia.images import *  # NOQA
        >>> import wbia
        >>> ibs = wbia.opendb(defaultdb='testdb1')
        >>> gsids = ibs._get_all_imgsetids()
        >>> self = ImageSets(gsids, ibs)
        >>> print(self)
        <ImageSets(num=13)>

    """

    def __init__(self, gsids, ibs, config=None):
        super(ImageSets, self).__init__(gsids, ibs, config)

    @property
    def images(self):
        return [self._ibs.images(gids) for gids in self.gids]

    @property
    def annots(self):
        return [self._ibs.annots(aids) for aids in self.aids]



# -*- coding: utf-8 -*-
"""
Interface to SSD object proposals.
"""
import logging
import utool as ut
import vtool as vt
from six.moves import zip
from os.path import abspath, dirname, expanduser, join, exists  # NOQA
import numpy as np
import sys

(print, rrr, profile) = ut.inject2(__name__, '[ssd]')
logger = logging.getLogger('wbia')

# SCRIPT_PATH = abspath(dirname(__file__))
SCRIPT_PATH = abspath(expanduser(join('~', 'code', 'ssd')))

if not ut.get_argflag('--no-ssd'):
    try:
        assert exists(SCRIPT_PATH)

        def add_path(path):
            # if path not in sys.path:
            sys.path.insert(0, path)

        # Add pycaffe to PYTHONPATH
        pycaffe_path = join(SCRIPT_PATH, 'python')
        add_path(pycaffe_path)

        import caffe

        rrr(caffe)
        from google.protobuf import text_format
        from caffe.proto import caffe_pb2
    except AssertionError:
        logger.info('WARNING Failed to find ssd. ' 'SSD is unavailable')
        # if ut.SUPER_STRICT:
        #     raise
    except ImportError:
        logger.info('WARNING Failed to import caffe. ' 'SSD is unavailable')
        # if ut.SUPER_STRICT:
        #     raise


VERBOSE_SS = ut.get_argflag('--verbssd') or ut.VERBOSE


CONFIG_URL_DICT = {
    'pretrained-300-pascal': 'https://wildbookiarepository.azureedge.net/models/pretrained.ssd.300.pascal.prototxt',
    'pretrained-512-pascal': 'https://wildbookiarepository.azureedge.net/models/pretrained.ssd.512.pascal.prototxt',
    'pretrained-300-pascal-plus': 'https://wildbookiarepository.azureedge.net/models/pretrained.ssd.300.pascal.plus.prototxt',
    'pretrained-512-pascal-plus': 'https://wildbookiarepository.azureedge.net/models/pretrained.ssd.512.pascal.plus.prototxt',
    'pretrained-300-coco': 'https://wildbookiarepository.azureedge.net/models/pretrained.ssd.300.coco.prototxt',
    'pretrained-512-coco': 'https://wildbookiarepository.azureedge.net/models/pretrained.ssd.512.coco.prototxt',
    'pretrained-300-ilsvrc': 'https://wildbookiarepository.azureedge.net/models/pretrained.ssd.300.ilsvrc.prototxt',
    'pretrained-500-ilsvrc': 'https://wildbookiarepository.azureedge.net/models/pretrained.ssd.500.ilsvrc.prototxt',
    'default': 'https://wildbookiarepository.azureedge.net/models/pretrained.ssd.512.pascal.plus.prototxt',
    None: 'https://wildbookiarepository.azureedge.net/models/pretrained.ssd.512.pascal.plus.prototxt',
}


def _parse_weight_from_cfg(url):
    return url.replace('.prototxt', '.caffemodel')


def _parse_classes_from_cfg(url):
    return url.replace('.prototxt', '.classes')


def _parse_class_list(classes_filepath):
    # Load classes from file into the class list
    assert exists(classes_filepath)
    class_list = []
    with open(classes_filepath) as classes:
        for line in classes.readlines():
            line = line.strip()
            if len(line) > 0:
                class_list.append(line)
    return class_list


def detect_gid_list(ibs, gid_list, downsample=True, verbose=VERBOSE_SS, **kwargs):
    """
    Args:
        gid_list (list of int): the list of IBEIS image_rowids that need detection
        downsample (bool, optional): a flag to indicate if the original image
            sizes should be used; defaults to True

            True:  ibs.get_image_detectpaths() is used
            False: ibs.get_image_paths() is used

    Kwargs (optional): refer to the SSD documentation for configuration settings

    Args:
        ibs (wbia.IBEISController):  image analysis api
        gid_list (list of int): the list of IBEIS image_rowids that need detection
        downsample (bool, optional): a flag to indicate if the original image
                sizes should be used; defaults to True

    Kwargs:
        detector, config_filepath, weights_filepath, verbose

    Yields:
        tuple: (gid, gpath, result_list)

    CommandLine:
        python -m wbia.algo.detect.ssd detect_gid_list --show

    Example:
        >>> # DISABLE_DOCTEST
        >>> from wbia.algo.detect.ssd import *  # NOQA
        >>> from wbia.core_images import LocalizerConfig
        >>> import wbia
        >>> ibs = wbia.opendb('testdb1')
        >>> gid_list = ibs.get_valid_gids()
        >>> config = {'verbose': True}
        >>> downsample = False
        >>> results_list = detect_gid_list(ibs, gid_list, downsample, **config)
        >>> results_list = list(results_list)
        >>> print('result lens = %r' % (map(len, list(results_list))))
        >>> print('result[0] = %r' % (len(list(results_list[0][2]))))
        >>> config = {'verbose': True}
        >>> downsample = False
        >>> results_list = detect_gid_list(ibs, gid_list, downsample, **config)
        >>> results_list = list(results_list)
        >>> print('result lens = %r' % (map(len, list(results_list))))
        >>> print('result[0] = %r' % (len(list(results_list[0][2]))))
        >>> ut.quit_if_noshow()
        >>> import wbia.plottool as pt
        >>> ut.show_if_requested()

    Yields:
        results (list of dict)
    """
    # Get new gpaths if downsampling
    if downsample:
        gpath_list = ibs.get_image_detectpaths(gid_list)
        neww_list = [vt.open_image_size(gpath)[0] for gpath in gpath_list]
        oldw_list = [oldw for (oldw, oldh) in ibs.get_image_sizes(gid_list)]
        downsample_list = [oldw / neww for oldw, neww in zip(oldw_list, neww_list)]
        orient_list = [1] * len(gid_list)
    else:
        gpath_list = ibs.get_image_paths(gid_list)
        downsample_list = [None] * len(gpath_list)
        orient_list = ibs.get_image_orientation(gid_list)
    # Run detection
    results_iter = detect(gpath_list, verbose=verbose, **kwargs)
    # Upscale the results
    _iter = zip(downsample_list, gid_list, orient_list, results_iter)
    for downsample, gid, orient, (gpath, result_list) in _iter:
        # Upscale the results back up to the original image size
        for result in result_list:
            if downsample is not None and downsample != 1.0:
                for key in ['xtl', 'ytl', 'width', 'height']:
                    result[key] = int(result[key] * downsample)
            bbox = (
                result['xtl'],
                result['ytl'],
                result['width'],
                result['height'],
            )
            bbox_list = [bbox]
            bbox = bbox_list[0]
            result['xtl'], result['ytl'], result['width'], result['height'] = bbox
        yield (gid, gpath, result_list)


def detect(
    gpath_list,
    config_filepath,
    weight_filepath,
    class_filepath,
    sensitivity,
    verbose=VERBOSE_SS,
    use_gpu=True,
    use_gpu_id=0,
    **kwargs,
):
    """
    Args:
        gpath_list (list of str): the list of image paths that need proposal candidates

    Kwargs (optional): refer to the SSD documentation for configuration settings

    Returns:
        iter
    """

    def _get_label_name(class_labelmap, label_list):
        if not isinstance(label_list, list):
            label_list = [label_list]
        item_list = class_labelmap.item
        name_list = []
        for label in label_list:
            found = False
            for i in range(len(item_list)):
                if label == item_list[i].label:
                    found = True
                    name_list.append(item_list[i].display_name)
                    break
            assert found
        return name_list

    # Get correct config if specified with shorthand
    config_url = None
    if config_filepath in CONFIG_URL_DICT:
        config_url = CONFIG_URL_DICT[config_filepath]
        config_filepath = ut.grab_file_url(config_url, appname='wbia', check_hash=True)

    # Get correct weights if specified with shorthand
    if weight_filepath in CONFIG_URL_DICT:
        if weight_filepath is None and config_url is not None:
            config_url_ = config_url
        else:
            config_url_ = CONFIG_URL_DICT[weight_filepath]
        weight_url = _parse_weight_from_cfg(config_url_)
        weight_filepath = ut.grab_file_url(weight_url, appname='wbia', check_hash=True)

    if class_filepath is None:
        class_url = _parse_classes_from_cfg(config_url)
        class_filepath = ut.grab_file_url(
            class_url, appname='wbia', check_hash=True, verbose=verbose
        )

    # load class labels
    with open(class_filepath, 'r') as class_file:
        class_labelmap = caffe_pb2.LabelMap()
        class_str = str(class_file.read(class_file))
        text_format.Merge(class_str, class_labelmap)

    # Need to convert unicode strings to Python strings to support Boost Python
    # call signatures in caffe
    prototxt_filepath = str(config_filepath)  # alias to Caffe nomenclature
    caffemodel_filepath = str(weight_filepath)  # alias to Caffe nomenclature

    assert exists(prototxt_filepath), 'Specified prototxt file not found'
    assert exists(caffemodel_filepath), 'Specified caffemodel file not found'

    if use_gpu:
        caffe.set_mode_gpu()
        caffe.set_device(use_gpu_id)
    else:
        caffe.set_mode_cpu()

    net = caffe.Net(prototxt_filepath, caffemodel_filepath, caffe.TEST)

    # Determine input size from prototext
    with open(prototxt_filepath, 'r') as prototxt_file:
        # load all lines
        line_list = prototxt_file.readlines()
        # look for dim size lines
        line_list = [line for line in line_list if 'dim:' in line]
        line_list = line_list[:4]
        # Get last line
        line = line_list[-1]
        line_ = line.strip().split(' ')
        # Filter empty spaces
        line_ = [_ for _ in line_ if len(_) > 0]
        # Get last value on line, which should be the image size
        image_resize = int(line_[-1])
        # Check to make sure
        assert image_resize in [300, 500, 512]
        logger.info('FOUND image_resize = %r' % (image_resize,))

    # Input preprocessing: 'data' is the name of the input blob == net.inputs[0]
    transformer = caffe.io.Transformer({'data': net.blobs['data'].data.shape})
    transformer.set_transpose('data', (2, 0, 1))
    # Mean pixel value
    transformer.set_mean('data', np.array([104, 117, 123]))
    # The reference model operates on images in [0,255] range instead of [0,1]
    transformer.set_raw_scale('data', 255)
    # The reference model has channels in BGR order instead of RGB
    transformer.set_channel_swap('data', (2, 1, 0))
    # Set batch size to 1 and set testing image size
    net.blobs['data'].reshape(1, 3, image_resize, image_resize)

    results_list_ = []
    for gpath in gpath_list:
        image = caffe.io.load_image(gpath)
        transformed_image = transformer.preprocess('data', image)
        net.blobs['data'].data[...] = transformed_image

        # Forward pass.
        detections = net.forward()['detection_out']

        # Parse the outputs.
        det_label = detections[0, 0, :, 1]
        det_conf = detections[0, 0, :, 2]
        det_xmin = detections[0, 0, :, 3]
        det_ymin = detections[0, 0, :, 4]
        det_xmax = detections[0, 0, :, 5]
        det_ymax = detections[0, 0, :, 6]

        # Get detections with confidence higher than 0.6.
        top_indices = [i for i, conf in enumerate(det_conf) if conf >= sensitivity]
        top_conf = det_conf[top_indices]
        top_label_indices = det_label[top_indices].tolist()
        top_labels = _get_label_name(class_labelmap, top_label_indices)
        top_xmin = det_xmin[top_indices]
        top_ymin = det_ymin[top_indices]
        top_xmax = det_xmax[top_indices]
        top_ymax = det_ymax[top_indices]
        height, width = image.shape[:2]

        # Compile results
        result_list_ = []
        zipped = zip(top_xmin, top_ymin, top_xmax, top_ymax, top_labels, top_conf)
        for (xmin, ymin, xmax, ymax, label, conf) in zipped:
            xtl = int(np.around(xmin * width))
            ytl = int(np.around(ymin * height))
            xbr = int(np.around(xmax * width))
            ybr = int(np.around(ymax * height))
            confidence = float(conf)
            result_dict = {
                'xtl': xtl,
                'ytl': ytl,
                'width': xbr - xtl,
                'height': ybr - ytl,
                'class': label,
                'confidence': confidence,
            }
            result_list_.append(result_dict)
        results_list_.append(result_list_)

    results_list = zip(gpath_list, results_list_)
    return results_list



# -*- coding: utf-8 -*-
"""
Runs functions in pipeline to get query reuslts and does some caching.
"""
import logging
import ubelt as ub
import utool as ut
from os.path import exists
from wbia.algo.hots import chip_match
from wbia.algo.hots import pipeline

(print, rrr, profile) = ut.inject2(__name__)
logger = logging.getLogger('wbia')


# TODO: Move to params
USE_HOTSPOTTER_CACHE = pipeline.USE_HOTSPOTTER_CACHE
USE_CACHE = not ut.get_argflag(('--nocache-query', '--noqcache')) and USE_HOTSPOTTER_CACHE
USE_BIGCACHE = (
    not ut.get_argflag(
        ('--nocache-big', '--no-bigcache-query', '--noqcache', '--nobigcache')
    )
    and ut.USE_CACHE
)
USE_SUPERCACHE = (
    not ut.get_argflag(
        ('--nocache-super', '--no-supercache-query', '--noqcache', '--nosupercache')
    )
    and ut.USE_CACHE
)
SAVE_CACHE = not ut.get_argflag('--nocache-save')
# MIN_BIGCACHE_BUNDLE = 20
# MIN_BIGCACHE_BUNDLE = 150
MIN_BIGCACHE_BUNDLE = 64
HOTS_BATCH_SIZE = ut.get_argval('--hots-batch-size', type_=int, default=None)


# ----------------------
# Main Query Logic
# ----------------------


@profile
def submit_query_request(
    qreq_,
    use_cache=None,
    use_bigcache=None,
    verbose=None,
    save_qcache=None,
    use_supercache=None,
    invalidate_supercache=None,
):
    """
    Called from qreq_.execute

    Checks a big cache for qaid2_cm.  If cache miss, tries to load each cm
    individually.  On an individual cache miss, it preforms the query.

    CommandLine:
        python -m wbia.algo.hots.match_chips4 --test-submit_query_request

    Examples:
        >>> # SLOW_DOCTEST
        >>> # xdoctest: +SKIP
        >>> from wbia.algo.hots.match_chips4 import *  # NOQA
        >>> import wbia
        >>> qaid_list = [1]
        >>> daid_list = [1, 2, 3, 4, 5]
        >>> use_bigcache = True
        >>> use_cache = True
        >>> ibs = wbia.opendb(db='testdb1')
        >>> qreq_ = ibs.new_query_request(qaid_list, daid_list, verbose=True)
        >>> cm_list = submit_query_request(qreq_=qreq_)
    """
    # Get flag defaults if necessary
    if verbose is None:
        verbose = pipeline.VERB_PIPELINE
    if use_cache is None:
        use_cache = USE_CACHE
    if save_qcache is None:
        save_qcache = SAVE_CACHE
    if use_bigcache is None:
        use_bigcache = USE_BIGCACHE
    if use_supercache is None:
        use_supercache = USE_SUPERCACHE
    # Create new query request object to store temporary state
    if verbose:
        # logger.info('[mc4] --- Submit QueryRequest_ --- ')
        logger.info(ub.color_text('[mc4] --- Submit QueryRequest_ --- ', 'yellow'))
    assert qreq_ is not None, 'query request must be prebuilt'

    # Check fo empty queries
    try:
        assert len(qreq_.daids) > 0, 'there are no database chips'
        assert len(qreq_.qaids) > 0, 'there are no query chips'
    except AssertionError as ex:
        ut.printex(
            ex,
            'Impossible query request',
            iswarning=True,
            keys=['qreq_.qaids', 'qreq_.daids'],
        )
        if ut.SUPER_STRICT:
            raise
        cm_list = [None for qaid in qreq_.qaids]
    else:
        # --- BIG CACHE ---
        # Do not use bigcache single queries
        is_big = len(qreq_.qaids) > MIN_BIGCACHE_BUNDLE
        use_bigcache_ = use_bigcache and use_cache and is_big
        if use_bigcache_ or save_qcache:
            cacher = qreq_.get_big_cacher()
            if use_bigcache_:
                try:
                    qaid2_cm = cacher.load()
                    cm_list = [qaid2_cm[qaid] for qaid in qreq_.qaids]
                except (IOError, AttributeError):
                    pass
                else:
                    return cm_list
        # ------------
        # Execute query request
        qaid2_cm = execute_query_and_save_L1(
            qreq_,
            use_cache,
            save_qcache,
            verbose=verbose,
            use_supercache=use_supercache,
            invalidate_supercache=invalidate_supercache,
        )
        # ------------
        if save_qcache and is_big:
            cacher.save(qaid2_cm)

        cm_list = [qaid2_cm[qaid] for qaid in qreq_.qaids]
    return cm_list


@profile
def execute_query_and_save_L1(
    qreq_,
    use_cache,
    save_qcache,
    verbose=True,
    batch_size=None,
    use_supercache=False,
    invalidate_supercache=False,
):
    """
    Args:
        qreq_ (wbia.QueryRequest):
        use_cache (bool):

    Returns:
        qaid2_cm

    CommandLine:
        python -m wbia.algo.hots.match_chips4 execute_query_and_save_L1:0
        python -m wbia.algo.hots.match_chips4 execute_query_and_save_L1:1
        python -m wbia.algo.hots.match_chips4 execute_query_and_save_L1:2
        python -m wbia.algo.hots.match_chips4 execute_query_and_save_L1:3


    Example0:
        >>> # SLOW_DOCTEST
        >>> # xdoctest: +SKIP
        >>> from wbia.algo.hots.match_chips4 import *  # NOQA
        >>> cfgdict1 = dict(codename='vsmany', sv_on=True)
        >>> p = 'default' + ut.get_cfg_lbl(cfgdict1)
        >>> qreq_ = wbia.main_helpers.testdata_qreq_(p=p, qaid_override=[1, 2, 3, 4)
        >>> use_cache, save_qcache, verbose = False, False, True
        >>> qaid2_cm = execute_query_and_save_L1(qreq_, use_cache, save_qcache, verbose)
        >>> print(qaid2_cm)

    Example1:
        >>> # SLOW_DOCTEST
        >>> # xdoctest: +SKIP
        >>> from wbia.algo.hots.match_chips4 import *  # NOQA
        >>> cfgdict1 = dict(codename='vsone', sv_on=True)
        >>> p = 'default' + ut.get_cfg_lbl(cfgdict1)
        >>> qreq_ = wbia.main_helpers.testdata_qreq_(p=p, qaid_override=[1, 2, 3, 4)
        >>> use_cache, save_qcache, verbose = False, False, True
        >>> qaid2_cm = execute_query_and_save_L1(qreq_, use_cache, save_qcache, verbose)
        >>> print(qaid2_cm)

    Example1:
        >>> # SLOW_DOCTEST
        >>> # xdoctest: +SKIP
        >>> # TEST SAVE
        >>> from wbia.algo.hots.match_chips4 import *  # NOQA
        >>> import wbia
        >>> cfgdict1 = dict(codename='vsmany', sv_on=True)
        >>> p = 'default' + ut.get_cfg_lbl(cfgdict1)
        >>> qreq_ = wbia.main_helpers.testdata_qreq_(p=p, qaid_override=[1, 2, 3, 4)
        >>> use_cache, save_qcache, verbose = False, True, True
        >>> qaid2_cm = execute_query_and_save_L1(qreq_, use_cache, save_qcache, verbose)
        >>> print(qaid2_cm)

    Example2:
        >>> # SLOW_DOCTEST
        >>> # xdoctest: +SKIP
        >>> # TEST LOAD
        >>> from wbia.algo.hots.match_chips4 import *  # NOQA
        >>> import wbia
        >>> cfgdict1 = dict(codename='vsmany', sv_on=True)
        >>> p = 'default' + ut.get_cfg_lbl(cfgdict1)
        >>> qreq_ = wbia.main_helpers.testdata_qreq_(p=p, qaid_override=[1, 2, 3, 4)
        >>> use_cache, save_qcache, verbose = True, True, True
        >>> qaid2_cm = execute_query_and_save_L1(qreq_, use_cache, save_qcache, verbose)
        >>> print(qaid2_cm)

    Example2:
        >>> # ENABLE_DOCTEST
        >>> # TEST PARTIAL HIT
        >>> from wbia.algo.hots.match_chips4 import *  # NOQA
        >>> import wbia
        >>> cfgdict1 = dict(codename='vsmany', sv_on=False, prescore_method='csum')
        >>> p = 'default' + ut.get_cfg_lbl(cfgdict1)
        >>> qreq_ = wbia.main_helpers.testdata_qreq_(p=p, qaid_override=[1, 2, 3,
        >>>                                                               4, 5, 6,
        >>>                                                               7, 8, 9])
        >>> use_cache, save_qcache, verbose = False, True, False
        >>> qaid2_cm = execute_query_and_save_L1(qreq_, use_cache,
        >>>                                      save_qcache, verbose,
        >>>                                      batch_size=3)
        >>> cm = qaid2_cm[1]
        >>> ut.delete(cm.get_fpath(qreq_))
        >>> cm = qaid2_cm[4]
        >>> ut.delete(cm.get_fpath(qreq_))
        >>> cm = qaid2_cm[5]
        >>> ut.delete(cm.get_fpath(qreq_))
        >>> cm = qaid2_cm[6]
        >>> ut.delete(cm.get_fpath(qreq_))
        >>> print('Re-execute')
        >>> qaid2_cm_ = execute_query_and_save_L1(qreq_, use_cache,
        >>>                                       save_qcache, verbose,
        >>>                                       batch_size=3)
        >>> assert all([qaid2_cm_[qaid] == qaid2_cm[qaid] for qaid in qreq_.qaids])
        >>> [ut.delete(fpath) for fpath in qreq_.get_chipmatch_fpaths(qreq_.qaids)]

    Ignore:
        other = cm_ = qaid2_cm_[qaid]
        cm = qaid2_cm[qaid]
    """
    if invalidate_supercache:
        dpath = qreq_.get_qresdir()
        fpath_list = ut.glob('%s/*_cm_supercache_*' % (dpath,))
        for fpath in fpath_list:
            ut.delete(fpath)

    if use_cache:
        if verbose:
            logger.info('[mc4] cache-query is on')
        if use_supercache:
            logger.info('[mc4] supercache-query is on')
        # Try loading as many cached results as possible
        qaid2_cm_hit = {}
        external_qaids = qreq_.qaids
        fpath_list = list(
            qreq_.get_chipmatch_fpaths(external_qaids, super_qres_cache=use_supercache)
        )
        exists_flags = [exists(fpath) for fpath in fpath_list]
        qaids_hit = ut.compress(external_qaids, exists_flags)
        fpaths_hit = ut.compress(fpath_list, exists_flags)
        fpath_iter = ut.ProgIter(
            fpaths_hit,
            length=len(fpaths_hit),
            enabled=len(fpaths_hit) > 1,
            label='loading cache hits',
            adjust=True,
            freq=1,
        )
        try:
            cm_hit_list = [
                chip_match.ChipMatch.load_from_fpath(fpath, verbose=False)
                for fpath in fpath_iter
            ]
            assert all(
                [qaid == cm.qaid for qaid, cm in zip(qaids_hit, cm_hit_list)]
            ), 'inconsistent qaid and cm.qaid'
            qaid2_cm_hit = {cm.qaid: cm for cm in cm_hit_list}
        except chip_match.NeedRecomputeError:
            logger.info('NeedRecomputeError: Some cached chips need to recompute')
            fpath_iter = ut.ProgIter(
                fpaths_hit,
                length=len(fpaths_hit),
                enabled=len(fpaths_hit) > 1,
                label='checking chipmatch cache',
                adjust=True,
                freq=1,
            )
            # Recompute those that fail loading
            qaid2_cm_hit = {}
            for fpath in fpath_iter:
                try:
                    cm = chip_match.ChipMatch.load_from_fpath(fpath, verbose=False)
                except chip_match.NeedRecomputeError:
                    pass
                else:
                    qaid2_cm_hit[cm.qaid] = cm
            logger.info(
                '%d / %d cached matches need to be recomputed'
                % (len(qaids_hit) - len(qaid2_cm_hit), len(qaids_hit))
            )
        if len(qaid2_cm_hit) == len(external_qaids):
            return qaid2_cm_hit
        else:
            if len(qaid2_cm_hit) > 0 and not ut.QUIET:
                logger.info(
                    '... partial cm cache hit %d/%d'
                    % (len(qaid2_cm_hit), len(external_qaids))
                )
        cachehit_qaids = list(qaid2_cm_hit.keys())
        # mask queries that have already been executed
        qreq_.set_external_qaid_mask(cachehit_qaids)
    else:
        if ut.VERBOSE:
            logger.info('[mc4] cache-query is off')
        qaid2_cm_hit = {}
    qaid2_cm = execute_query2(qreq_, verbose, save_qcache, batch_size, use_supercache)
    # Merge cache hits with computed misses
    if len(qaid2_cm_hit) > 0:
        qaid2_cm.update(qaid2_cm_hit)
    qreq_.set_external_qaid_mask(None)  # undo state changes
    return qaid2_cm


@profile
def execute_query2(qreq_, verbose, save_qcache, batch_size=None, use_supercache=False):
    """
    Breaks up query request into several subrequests
    to process "more efficiently" and safer as well.
    """
    if qreq_.prog_hook is not None:
        preload_hook, query_hook = qreq_.prog_hook.subdivide(spacing=[0, 0.15, 0.8])
        preload_hook(0, lbl='preloading')
        qreq_.prog_hook = query_hook
    else:
        preload_hook = None
    # Load features / weights for all annotations
    qreq_.lazy_preload(prog_hook=preload_hook, verbose=verbose and ut.NOT_QUIET)

    all_qaids = qreq_.qaids
    logger.info('len(missed_qaids) = %r' % (len(all_qaids),))
    qaid2_cm = {}
    # vsone must have a chunksize of 1
    if batch_size is None:
        if HOTS_BATCH_SIZE is None:
            hots_batch_size = qreq_.ibs.cfg.other_cfg.hots_batch_size
            # hots_batch_size = 256
        else:
            hots_batch_size = HOTS_BATCH_SIZE
    else:
        hots_batch_size = batch_size
    chunksize = 1 if qreq_.qparams.vsone else hots_batch_size

    # Iterate over vsone queries in chunks.
    n_total_chunks = ut.get_num_chunks(len(all_qaids), chunksize)
    qaid_chunk_iter = ut.ichunks(all_qaids, chunksize)
    _qreq_iter = (qreq_.shallowcopy(qaids=qaids) for qaids in qaid_chunk_iter)
    sub_qreq_iter = ut.ProgIter(
        _qreq_iter,
        length=n_total_chunks,
        freq=1,
        label='[mc4] query chunk: ',
        prog_hook=qreq_.prog_hook,
    )
    for sub_qreq_ in sub_qreq_iter:
        if ut.VERBOSE:
            logger.info('Generating vsmany chunk')
        sub_cm_list = pipeline.request_wbia_query_L0(
            qreq_.ibs, sub_qreq_, verbose=verbose
        )
        assert len(sub_qreq_.qaids) == len(sub_cm_list), 'not aligned'
        assert all(
            [qaid == cm.qaid for qaid, cm in zip(sub_qreq_.qaids, sub_cm_list)]
        ), 'not corresonding'
        if save_qcache:
            fpath_list = list(
                qreq_.get_chipmatch_fpaths(
                    sub_qreq_.qaids, super_qres_cache=use_supercache
                )
            )
            _iter = zip(sub_cm_list, fpath_list)
            _iter = ut.ProgIter(
                _iter,
                length=len(sub_cm_list),
                label='saving chip matches',
                adjust=True,
                freq=1,
            )
            for cm, fpath in _iter:
                cm.save_to_fpath(fpath, verbose=False)
        else:
            if ut.VERBOSE:
                logger.info('[mc4] not saving vsmany chunk')
        qaid2_cm.update({cm.qaid: cm for cm in sub_cm_list})
    return qaid2_cm



# -*- coding: utf-8 -*-
import logging
import utool as ut
import numpy as np  # NOQA
from six.moves import range

(print, rrr, profile) = ut.inject2(__name__, '[_wbia_object]')
logger = logging.getLogger('wbia')


def _find_wbia_attrs(ibs, objname, blacklist=[]):
    r"""
    Developer function to help figure out what attributes are available

    Args:
        ibs (wbia.IBEISController):  images analysis api

    CommandLine:
        python -m wbia.images _find_wbia_attrs

    Example:
        >>> # DISABLE_DOCTEST
        >>> from wbia._wbia_object import *  # NOQA
        >>> import wbia
        >>> ibs = wbia.opendb(defaultdb='testdb1')
        >>> objname = 'images'
        >>> blacklist = []
        >>> _find_wbia_attrs(ibs, objname, blacklist)

    Example:
        >>> # DISABLE_DOCTEST
        >>> from wbia._wbia_object import *  # NOQA
        >>> import wbia
        >>> ibs = wbia.opendb(defaultdb='testdb1')
        >>> objname = 'annot'
        >>> blacklist = ['annot_pair']
        >>> _find_wbia_attrs(ibs, objname, blacklist)
    """
    import re

    getter_prefix = 'get_' + objname + '_'
    found_getters = ut.search_module(ibs, getter_prefix)
    pat = getter_prefix + ut.named_field('attr', '.*')
    for stopword in blacklist:
        found_getters = [fn for fn in found_getters if stopword not in fn]
    matched_getters = [re.match(pat, fn).groupdict()['attr'] for fn in found_getters]

    setter_prefix = 'set_' + objname + '_'
    found_setters = ut.search_module(ibs, setter_prefix)
    pat = setter_prefix + ut.named_field('attr', '.*')
    for stopword in blacklist:
        found_setters = [fn for fn in found_setters if stopword not in fn]
    matched_setters = [re.match(pat, fn).groupdict()['attr'] for fn in found_setters]
    return matched_getters, matched_setters


def _inject_getter_attrs(
    metaself,
    objname,
    attrs,
    configurable_attrs,
    depc_name=None,
    depcache_attrs=None,
    settable_attrs=None,
    aliased_attrs=None,
):
    """
    Used by the metaclass to inject methods and properties into the class
    inheriting from ObjectList1D
    """

    if settable_attrs is None:
        settable_attrs = []
    settable_attrs = set(settable_attrs)

    # Inform the class of which variables will be injected
    metaself._settable_attrs = settable_attrs
    metaself._attrs = attrs
    metaself._configurable_attrs = configurable_attrs
    if depcache_attrs is None:
        metaself._depcache_attrs = []
    else:
        metaself._depcache_attrs = ['%s_%s' % (tbl, col) for tbl, col in depcache_attrs]
    if aliased_attrs is not None:
        metaself._attrs_aliases = aliased_attrs
    else:
        metaself._attrs_aliases = {}

    # if not getattr(metaself, '__needs_inject__', True):
    #     return

    attr_to_aliases = ut.invert_dict(metaself._attrs_aliases, unique_vals=False)

    # What is difference between configurable and depcache getters?
    # Could depcache getters just be made configurable?
    # I guess its just an efficincy thing. Actually its config2_-vs-config
    # FIXME: rectify differences between normal / configurable / depcache
    # getter

    def _make_caching_setter(attrname, _rowid_setter):
        def _setter(self, values, *args, **kwargs):
            if self._ibs is None:
                self._internal_attrs[attrname] = values
            else:
                if self._caching and attrname in self._internal_attrs:
                    self._internal_attrs[attrname] = values
                _rowid_setter(self, self._rowids, values)

        ut.set_funcname(_setter, '_set_' + attrname)
        return _setter

    def _make_caching_getter(attrname, _rowid_getter):
        def _getter(self):
            if self._ibs is None or (self._caching and attrname in self._internal_attrs):
                data = self._internal_attrs[attrname]
            else:
                data = _rowid_getter(self, self._rowids)
                if self._caching:
                    self._internal_attrs[attrname] = data
            return data

        ut.set_funcname(_getter, '_get_' + attrname)
        return _getter

    # make default version use implicit rowids and another
    # that takes explicit rowids.

    def _make_setters(objname, attrname):
        ibs_funcname = 'set_%s_%s' % (objname, attrname)

        def _rowid_setter(self, rowids, values, *args, **kwargs):
            ibs_callable = getattr(self._ibs, ibs_funcname)
            ibs_callable(rowids, values, *args, **kwargs)

        ut.set_funcname(_rowid_setter, '_rowid_set_' + attrname)
        _setter = _make_caching_setter(attrname, _rowid_setter)
        return _rowid_setter, _setter

    # ---

    def _make_getters(objname, attrname):
        ibs_funcname = 'get_%s_%s' % (objname, attrname)

        def _rowid_getter(self, rowids):
            ibs_callable = getattr(self._ibs, ibs_funcname)
            data = ibs_callable(rowids)
            if self._asarray:
                data = np.array(data)
            return data

        ut.set_funcname(_rowid_getter, '_rowid_get_' + attrname)
        _getter = _make_caching_getter(attrname, _rowid_getter)
        return _rowid_getter, _getter

    def _make_cfg_getters(objname, attrname):
        ibs_funcname = 'get_%s_%s' % (objname, attrname)

        def _rowid_getter(self, rowids):
            ibs_callable = getattr(self._ibs, ibs_funcname)
            data = ibs_callable(rowids, config2_=self._config)
            if self._asarray:
                data = np.array(data)
            return data

        ut.set_funcname(_rowid_getter, '_rowid_get_' + attrname)
        _getter = _make_caching_getter(attrname, _rowid_getter)
        return _rowid_getter, _getter

    def _make_depc_getters(depc_name, attrname, tbl, col):
        def _rowid_getter(self, rowids):
            depc = getattr(self._ibs, depc_name)
            data = depc.get(tbl, rowids, col, config=self._config)
            if self._asarray:
                data = np.array(data)
            return data

        ut.set_funcname(_rowid_getter, '_rowid_get_' + attrname)
        _getter = _make_caching_getter(attrname, _rowid_getter)
        return _rowid_getter, _getter

    # Collect setter / getter functions and properties
    rowid_getters = []
    getters = []
    setters = []
    properties = []
    for attrname in attrs:
        _rowid_getter, _getter = _make_getters(objname, attrname)
        if attrname in settable_attrs:
            _rowid_setter, _setter = _make_setters(objname, attrname)
            setters.append(_setter)
        else:
            _setter = None
        prop = property(fget=_getter, fset=_setter)
        rowid_getters.append((attrname, _rowid_getter))
        getters.append(_getter)
        properties.append((attrname, prop))

    for attrname in configurable_attrs:
        _rowid_getter, _getter = _make_cfg_getters(objname, attrname)
        prop = property(fget=_getter)
        rowid_getters.append((attrname, _rowid_getter))
        getters.append(_getter)
        properties.append((attrname, prop))

    if depcache_attrs is not None:
        for tbl, col in depcache_attrs:
            attrname = '%s_%s' % (tbl, col)
            _rowid_getter, _getter = _make_depc_getters(depc_name, attrname, tbl, col)
            prop = property(fget=_getter, fset=None)
            rowid_getters.append((attrname, _rowid_getter))
            getters.append(_getter)
            properties.append((attrname, prop))

    aliases = []

    # Inject all gathered information
    for attrname, func in rowid_getters:
        funcname = ut.get_funcname(func)
        setattr(metaself, funcname, func)
        # ensure aliases have rowid getters
        for alias in attr_to_aliases.get(attrname, []):
            alias_funcname = '_rowid_get_' + alias
            setattr(metaself, alias_funcname, func)

    for func in getters:
        funcname = ut.get_funcname(func)
        setattr(metaself, funcname, func)

    for func in setters:
        funcname = ut.get_funcname(func)
        setattr(metaself, funcname, func)

    for attrname, prop in properties:
        setattr(metaself, attrname, prop)
        for alias in attr_to_aliases.pop(attrname, []):
            aliases.append((alias, attrname))
            setattr(metaself, alias, prop)

    if ut.get_argflag('--autogen-core'):
        # TODO: turn on autogenertion given a flag
        def expand_closure_source(funcname, func):
            source = ut.get_func_sourcecode(func)
            closure_vars = [
                (k, v.cell_contents)
                for k, v in zip(func.func_code.co_freevars, func.func_closure)
            ]
            source = ut.unindent(source)
            import re

            for k, v in closure_vars:
                source = re.sub('\\b' + k + '\\b', ut.repr2(v), source)
            source = re.sub(r'def .*\(self', 'def ' + funcname + '(self', source)
            source = ut.indent(source.strip(), '    ') + '\n'
            return source

        explicit_lines = []
        # build explicit version for jedi?
        for funcname, func in getters:
            source = expand_closure_source(funcname, func)
            explicit_lines.append(source)
        # build explicit version for jedi?
        for funcname, func in setters:
            source = expand_closure_source(funcname, func)
            explicit_lines.append(source)

        for attrname, prop in properties:
            getter_name = None if prop.fget is None else ut.get_funcname(prop.fget)
            setter_name = None if prop.fset is None else ut.get_funcname(prop.fset)
            source = '    %s = property(%s, %s)' % (attrname, getter_name, setter_name)
            explicit_lines.append(source)

        for alias, attrname in aliases:
            source = '    %s = %s' % (alias, attrname)
            explicit_lines.append(source)

        explicit_source = (
            '\n'.join(
                [
                    'from wbia import _wbia_object',
                    '',
                    '',
                    'class _%s_base_class(_wbia_object.ObjectList1D):',
                    '    __needs_inject__ = False',
                    '',
                ]
            )
            % (objname,)
        )
        explicit_source += '\n'.join(explicit_lines)
        explicit_fname = '_autogen_%s_base.py' % (objname,)
        from os.path import dirname, join

        ut.writeto(join(dirname(__file__), explicit_fname), explicit_source + '\n')

    if attr_to_aliases:
        raise AssertionError('Unmapped aliases %r' % (attr_to_aliases,))


class ObjectScalar0D(ut.NiceRepr, ut.HashComparable2):
    """
    This actually stores a ObjectList1D of length 1 and
    simply calls those functions where available
    """

    def __init__(self, obj1d):
        assert len(obj1d) == 1
        self.obj1d = obj1d

    def __nice__(self):
        return 'rowid=%s, uuid=%s' % (self._rowids, self.uuids)

    def __getattr__(self, key):
        vals = getattr(self.obj1d, key)
        if key == 'show':
            return vals
        return vals[0]

    def __dir__(self):
        attrs = dir(object)
        attrs += list(self.__class__.__dict__.keys())
        attrs += self.obj1d.__vector_attributes__()
        return attrs

    def _make_lazy_dict(self):
        """
        CommandLine:
            python -m wbia._wbia_object ObjectScalar0D._make_lazy_dict

        Example:
            >>> # DISABLE_DOCTEST
            >>> from wbia._wbia_object import *  # NOQA
            >>> import wbia
            >>> ibs = wbia.opendb('testdb1')
            >>> annots = ibs.annots()
            >>> subset = annots.take([0, 2, 5])
            >>> scalar = annots[0]
            >>> assert scalar.obj1d._attrs == annots._attrs
            >>> self = scalar
            >>> print(dir(self))
            >>> metadata = self._make_lazy_dict()
            >>> print('metadata = %r' % (metadata,))
            >>> aid = metadata['aid']
            >>> print('aid = %r' % (aid,))
        """
        metadata = ut.LazyDict()
        for attr in self.obj1d.__vector_attributes__():
            metadata[attr] = ut.partial(getattr, self, attr)
        return metadata


# @ut.reloadable_class
class ObjectList1D(ut.NiceRepr, ut.HashComparable2):
    """
    An object that efficiently operates on a list of wbia objects using
    vectorized code. Single instances can be returned as ObjectScalar0D's
    """

    def __init__(self, rowids, ibs, config=None, caching=False, asarray=False):
        self._rowids = rowids
        # self._islist = True
        # Internal cache
        self._internal_attrs = {}
        # Internal behaviors
        self._ibs = ibs
        self._config = config
        self._caching = caching
        # Private attributes
        self._rowid_to_idx = None
        self._asarray = asarray
        # ut.make_index_lookup(self._rowids)

    def __vector_attributes__(self):
        attrs = (
            self._attrs
            + self._configurable_attrs
            + self._depcache_attrs
            + list(self._attrs_aliases.keys())
        )
        return attrs

    def set_caching(self, flag):
        self._caching = flag

    def __nice__(self):
        return 'num=%r' % (len(self))

    def __hash__(self):
        return hash(self.group_uuid())

    def __add__(self, other):
        assert self.__class__ is other.__class__, 'incompatable'
        assert self._ibs is other._ibs, 'incompatable'
        assert self._config is other._config, 'incompatable'
        rowids = ut.unique(self._rowids + other._rowids)
        new = self.__class__(rowids, self._ibs, self._config)
        return new

    def take(self, idxs):
        """
        Creates a subset of the list using the specified indices.
        """
        rowids = ut.take(self._rowids, idxs)
        # Create a new instance pointing only to the requested subset
        newself = self.__class__(
            rowids, ibs=self._ibs, config=self._config, caching=self._caching
        )
        # Pass along any internally cached values
        _new_internal = {
            key: ut.take(val, idxs) for key, val in self._internal_attrs.items()
        }
        newself._internal_attrs = _new_internal
        return newself

    def preload(self, *attrs):
        assert self._ibs is not None, 'must be connected to preload'
        for attrname in attrs:
            self._internal_attrs[attrname] = getattr(self, attrname)

    def group_uuid(self):
        sorted_uuids = sorted(self.uuids)
        group_uuid = ut.util_hash.augment_uuid(*sorted_uuids)
        return group_uuid

    def disconnect(self):
        """
        Disconnects object from the state of the database. All information has
        been assumed to be preloaded.
        """
        self._ibs = None

    def __iter__(self):
        return iter(self._rowids)

    def __len__(self):
        return len(self._rowids)

    def __getitem__(self, idx):
        if isinstance(idx, slice):
            idxs = list(range(*idx.indices(len(self))))
            return self.take(idxs)
        if not ut.isiterable(idx):
            obj0d_ = self.take([idx])
            obj0d = ObjectScalar0D(obj0d_)
            return obj0d
        if not isinstance(idx, slice):
            raise AssertionError('only slice supported currently')
        return self.take(idx)

    def scalars(self):
        scalar_list = [self[idx] for idx in range(len(self))]
        return scalar_list

    def compress(self, flags):
        idxs = ut.where(flags)
        return self.take(idxs)

    def take_column(self, keys):
        vals_list = zip(*[getattr(self, key) for key in keys])
        dict_list = [dict(zip(keys, vals)) for vals in vals_list]
        return dict_list

    def chunks(self, chunksize):
        for idxs in ut.ichunks(self, range(len(self))):
            yield self.take(idxs)

    def group_indicies(self, labels):
        unique_labels, groupxs = ut.group_indices(labels)
        return unique_labels, groupxs

    def group_items(self, labels):
        """ group as dict """
        unique_labels, groups = self.group(labels)
        label_to_group = ut.odict(zip(unique_labels, groups))
        return label_to_group

    def group(self, labels):
        """ group as list """
        unique_labels, groupxs = self.group_indicies(labels)
        groups = [self.take(idxs) for idxs in groupxs]
        return unique_labels, groups

    def lookup_idxs(self, rowids):
        """ Lookup subset indicies by rowids """
        if self._rowid_to_idx is None:
            self._rowid_to_idx = ut.make_index_lookup(self._rowids)
        idx_list = ut.take(self._rowid_to_idx, rowids)
        return idx_list

    def loc(self, rowids):
        """ Lookup subset by rowids """
        idxs = self.lookup_idxs(rowids)
        return self.take(idxs)

    # def filter(self, filterkw):
    #     pass

    # def filter_flags(self, filterkw):
    #     pass

    def view(self, rowids=None):
        """
        Like take, but returns a view proxy that maps to the original parent
        """
        if rowids is None:
            rowids = self._rowids
        # unique_parent = self.take(unique_idxs)
        view = ObjectView1D(rowids, obj1d=self)
        return view


class ObjectView1D(ut.NiceRepr):
    # ut.HashComparable2):
    """
    Allows for proxy caching.

    Example:
        >>> # ENABLE_DOCTEST
        >>> from wbia._wbia_object import *  # NOQA
        >>> import wbia
        >>> ibs = wbia.opendb(defaultdb='testdb1')
        >>> aids = ibs.get_valid_aids()
        >>> a = self = annots = ibs.annots(aids)
        >>> rowids = [1, 1, 3, 2, 1, 2]
        >>> self = v = a.view(rowids)
        >>> assert np.all(v.vecs[0] == v.vecs[1])
        >>> assert v.vecs[0] is v.vecs[1]
        >>> assert v.vecs[0] is not v.vecs[2]
    """

    def __init__(self, rowids, obj1d, cache=None):
        self._rowids = list(rowids)
        self._obj1d = obj1d
        self._unique_rowids = set(self._rowids)
        self._unique_inverse = ut.list_alignment(self._unique_rowids, self._rowids)
        if cache is None:
            self._cache = ut.ddict(dict)
        else:
            self._cache = cache
        # Views always cache data for now
        self._caching = True

    def __dir__(self):
        attrs = dir(object)
        attrs += self.__dict__.keys()
        attrs += ['__dict__', '__module__', '__weakref__']
        # ['_unique_parent', '_caching', '_attr_rowid_value', '_rowids']
        attrs += list(self.__class__.__dict__.keys())
        attrs += self._obj1d.__vector_attributes__()
        return attrs

    def __vector_attributes__(self):
        return self._obj1d.__vector_attributes__()

    def __getattr__(self, key):
        """
        key = 'vecs'
        """
        try:
            _rowid_getter = getattr(self._obj1d, '_rowid_get_%s' % (key,))
        except AttributeError:
            raise AttributeError('ObjectView1D has no attribute %r' % (key,))
        if self._caching:
            rowid_to_value = self._cache[key]
            miss_rowids = [
                rowid for rowid in self._unique_rowids if rowid not in rowid_to_value
            ]
            miss_data = _rowid_getter(miss_rowids)
            for rowid, value in zip(miss_rowids, miss_data):
                rowid_to_value[rowid] = value
            unique_data = ut.take(rowid_to_value, self._unique_rowids)
        else:
            unique_data = _rowid_getter(self._unique_rowids)
        data = ut.take(unique_data, self._unique_inverse)
        return data

    def __iter__(self):
        return iter(self._rowids)

    def __len__(self):
        return len(self._rowids)

    def __nice__(self):
        return 'unique=%r, num=%r' % (len(self._unique_rowids), len(self))

    # def __hash__(self):
    #     return hash(self.group_uuid())

    def view(self, rowids):
        """
        returns a view of a view that uses the same per-item cache

        Example:
            >>> # ENABLE_DOCTEST
            >>> from wbia._wbia_object import *  # NOQA
            >>> import wbia
            >>> ibs = wbia.opendb(defaultdb='testdb1')
            >>> aids = ibs.get_valid_aids()
            >>> annots = ibs.annots(aids)
            >>> self = annots.view(annots._rowids)
            >>> v1 = self.view([1, 1, 2, 3, 1, 2])
            >>> v2 = self.view([3, 4, 5])
            >>> v3 = self.view([1, 4])
            >>> v4 = self.view(3)
            >>> lazy4 = v4._make_lazy_dict()
            >>> assert v1.vecs[0] is v3.vecs[0]
            >>> assert v2._cache is self._cache
            >>> assert v2._cache is v1._cache
        """
        if ut.isiterable(rowids):
            childview = self.__class__(rowids, obj1d=self._obj1d, cache=self._cache)
        else:
            childview = self.__class__([rowids], obj1d=self._obj1d, cache=self._cache)
            childview = ObjectScalar0D(childview)
        return childview



# -*- coding: utf-8 -*-
def custom_globals():
    import utool as ut

    ut.util_io.__PRINT_WRITES__ = False
    ut.util_io.__PRINT_READS__ = False
    ut.util_parallel.__FORCE_SERIAL__ = True
    ut.util_cache.VERBOSE_CACHE = False
    ut.NOT_QUIET = False

    import wbia.plottool as pt  # NOQA
    import matplotlib as mpl

    mpl.rcParams['axes.labelsize'] = 20
    mpl.rcParams['axes.titlesize'] = 20
    mpl.rcParams['figure.titlesize'] = 20


def make_cells_wider():
    # Make notebook cells wider
    from IPython.core.display import HTML

    # This must be the last line in a cell
    return HTML('<style>body .container { width:99% !important; }</style>')



# -*- coding: utf-8 -*-
"""
TODO: rewrite the hotspotter lnbnn algo to be a generator

Wrapper around LNBNN hotspotter algorithm
"""
# -*- coding: utf-8 -*-
import logging
import utool as ut

print, rrr, profile = ut.inject2(__name__)
logger = logging.getLogger('wbia')


class Ranker(object):
    def __init__(ranker, ibs=None, config={}):
        ranker.ibs = ibs
        ranker.config = config

        ranker._daids = None
        ranker._nids = None
        ranker.verbose = True

    def fit(ranker, daids, dnids=None):
        ranker._daids = daids
        ranker._nids = dnids
        pass

    def predict(ranker, qaids, qnids=None, prog_hook=None):
        custom_nid_lookup = dict(zip(ranker._daids, ranker._dnids))
        custom_nid_lookup.update(dict(zip(qaids, qnids)))

        cfgdict = ranker.config
        ibs = ranker.ibs
        qreq_ = ibs.new_query_request(
            qaids,
            ranker._daids,
            cfgdict=cfgdict,
            custom_nid_lookup=custom_nid_lookup,
            verbose=ranker.verbose,
        )
        cm_list = qreq_.execute(prog_hook=prog_hook)
        return cm_list

        # ibs = ranker.ibs
        # aids = sorted(set(ut.aslist(qaids) + ut.aslist(daids)))
        # custom_nid_lookup = infr.get_node_attrs('name_label', aids)
        # for qaid in qaids:
        #     pass



# -*- coding: utf-8 -*-
import utool as ut


__all__ = ('get_testdata_dir',)


def get_testdata_dir(ensure=True, key='testdb1'):
    """
    Gets test img directory and downloads it if it doesn't exist
    """
    testdata_map = {
        # 'testdb1': 'https://cthulhu.dyn.wildme.io/public/data/testdata.zip'}
        'testdb1': 'https://wildbookiarepository.azureedge.net/data/testdata.zip',
    }
    zipped_testdata_url = testdata_map[key]
    testdata_dir = ut.grab_zipped_url(zipped_testdata_url, ensure=ensure)
    return testdata_dir



# -*- coding: utf-8 -*-
"""
Hotspotter pipeline module

Module Notation and Concepts::
    PREFIXES:
    qaid2_XXX - prefix mapping query chip index to
    qfx2_XXX  - prefix mapping query chip feature index to

     * nns    - a (qfx2_idx, qfx2_dist) tuple

     * idx    - the index into the nnindexers descriptors
     * qfx    - query feature index wrt the query chip
     * dfx    - query feature index wrt the database chip
     * dist   - the distance to a corresponding feature
     * fm     - a list of feature match pairs / correspondences (qfx, dfx)
     * fsv    - a score vector of a corresponding feature
     * valid  - a valid bit for a corresponding feature

    PIPELINE_VARS::
    nns_list - maping from query chip index to nns
     * qfx2_idx   - ranked list of query feature indexes to database feature indexes
     * qfx2_dist - ranked list of query feature indexes to database feature indexes

    * qaid2_norm_weight - mapping from qaid to (qfx2_normweight, qfx2_selnorm)
             = qaid2_nnfiltagg[qaid]

CommandLine:
    To see the ouput of a complete pipeline run use

    # Set to whichever database you like
    python main.py --db PZ_MTEST --setdb
    python main.py --db NAUT_test --setdb
    python main.py --db testdb1 --setdb

    # Then run whichever configuration you like
    python main.py --query 1 --yes --noqcache -t default:codename=vsmany
    python main.py --query 1 --yes --noqcache -t default:codename=vsmany_nsum

TODO:
    * Don't preload the nn-indexer in case the nearest neighbors have already
    been computed?
"""
import logging
from six.moves import zip, range, map
import numpy as np
import vtool as vt
from wbia.algo.hots import hstypes
from wbia.algo.hots import chip_match
from wbia.algo.hots import nn_weights
from wbia.algo.hots import scoring
from wbia.algo.hots import _pipeline_helpers as plh  # NOQA
from collections import namedtuple
import utool as ut

print, rrr, profile = ut.inject2(__name__)
logger = logging.getLogger('wbia')


# =================
# Globals
# =================

TAU = 2 * np.pi  # References: tauday.com
NOT_QUIET = ut.NOT_QUIET and not ut.get_argflag('--quiet-query')
DEBUG_PIPELINE = ut.get_argflag(('--debug-pipeline', '--debug-pipe'))
VERB_PIPELINE = NOT_QUIET and (
    ut.VERBOSE or ut.get_argflag(('--verbose-pipeline', '--verb-pipe'))
)
VERYVERBOSE_PIPELINE = ut.get_argflag(('--very-verbose-pipeline', '--very-verb-pipe'))

USE_HOTSPOTTER_CACHE = not ut.get_argflag('--nocache-hs') and ut.USE_CACHE
USE_NN_MID_CACHE = (
    (True and ut.is_developer())
    and not ut.get_argflag('--nocache-nnmid')
    and USE_HOTSPOTTER_CACHE
)
USE_NN_MID_CACHE = False


NN_LBL = 'Assign NN:       '
FILT_LBL = 'Filter NN:       '
WEIGHT_LBL = 'Weight NN:       '
BUILDCM_LBL = 'Build Chipmatch: '
SVER_LVL = 'SVER:            '

PROGKW = dict(freq=1, time_thresh=30.0, adjust=True)


# Internal tuples denoting return types
WeightRet_ = namedtuple(
    'weight_ret',
    ('filtkey_list', 'filtweights_list', 'filtvalids_list', 'filtnormks_list'),
)


class Neighbors(ut.NiceRepr):
    __slots__ = ['qaid', 'qfx_list', 'neighb_idxs', 'neighb_dists']

    # TODO: replace with named tuple?
    def __init__(self, qaid, idxs, dists, qfxs):
        self.qaid = qaid
        self.qfx_list = qfxs
        self.neighb_idxs = idxs
        self.neighb_dists = dists

    @property
    def num_query_feats(self):
        if self.qfx_list is None:
            return len(self.neighb_idxs)
        else:
            return len(self.qfx_list)

    def __iter__(self):
        return iter([self.neighb_idxs, self.neighb_dists])

    def __getitem__(self, index):
        return (self.neighb_idxs, self.neighb_dists)[index]

    def __nice__(self):
        return '(qaid=%r,nQfxs=%r,nNbs=%r)' % (
            self.qaid,
            self.num_query_feats,
            self.neighb_idxs.shape[1],
        )

    def __getstate__(self):
        return self.__dict__

    def __setstate__(self, state):
        return self.__dict__.update(**state)


# @profile
def request_wbia_query_L0(ibs, qreq_, verbose=VERB_PIPELINE):
    r"""Driver logic of query pipeline

    Note:
        Make sure _pipeline_helpres.testrun_pipeline_upto reflects what happens
        in this function.

    Args:
        ibs (wbia.IBEISController): IBEIS database object to be queried.
            technically this object already lives inside of qreq_.
        qreq_ (wbia.QueryRequest): hyper-parameters. use
            ``ibs.new_query_request`` to create one

    Returns:
        list: cm_list containing ``wbia.ChipMatch`` objects

    CommandLine:
        python -m wbia.algo.hots.pipeline --test-request_wbia_query_L0:0 --show
        python -m wbia.algo.hots.pipeline --test-request_wbia_query_L0:1 --show

        python -m wbia.algo.hots.pipeline --test-request_wbia_query_L0:0 --db testdb1 --qaid 325
        python -m wbia.algo.hots.pipeline --test-request_wbia_query_L0:0 --db testdb3 --qaid 325
        # background match
        python -m wbia.algo.hots.pipeline --test-request_wbia_query_L0:0 --db NNP_Master3 --qaid 12838

        python -m wbia.algo.hots.pipeline --test-request_wbia_query_L0:0
        python -m wbia.algo.hots.pipeline --test-request_wbia_query_L0:0 --db PZ_MTEST -a timectrl:qindex=0:256
        python    -m wbia.algo.hots.pipeline --test-request_wbia_query_L0:0 --db PZ_Master1 -a timectrl:qindex=0:256
        utprof.py -m wbia.algo.hots.pipeline --test-request_wbia_query_L0:0 --db PZ_Master1 -a timectrl:qindex=0:256

    Example1:
        >>> # ENABLE_DOCTEST
        >>> from wbia.algo.hots.pipeline import *  # NOQA
        >>> import wbia
        >>> qreq_ = wbia.init.main_helpers.testdata_qreq_(a=['default:qindex=0:2,dindex=0:10'])
        >>> ibs = qreq_.ibs
        >>> print(qreq_.qparams.query_cfgstr)
        >>> verbose = True
        >>> cm_list = request_wbia_query_L0(ibs, qreq_, verbose=verbose)
        >>> cm = cm_list[0]
        >>> ut.quit_if_noshow()
        >>> cm.ishow_analysis(qreq_, fnum=0, make_figtitle=True)
        >>> ut.show_if_requested()
    """
    # Load data for nearest neighbors
    if verbose:
        assert ibs is qreq_.ibs
        logger.info('\n\n[hs] +--- STARTING HOTSPOTTER PIPELINE ---')
        logger.info(ut.indent(qreq_.get_infostr(), '[hs] '))

    ibs.assert_valid_aids(qreq_.get_internal_qaids(), msg='pipeline qaids')
    ibs.assert_valid_aids(qreq_.get_internal_daids(), msg='pipeline daids')

    if qreq_.qparams.pipeline_root == 'smk':
        from wbia.algo.hots.smk import smk_match

        # Alternative to naive bayes matching:
        # Selective match kernel
        qaid2_scores, qaid2_chipmatch_FILT_ = smk_match.execute_smk_L5(qreq_)
    elif qreq_.qparams.pipeline_root in ['vsone', 'vsmany']:
        assert qreq_.qparams.pipeline_root != 'vsone', 'pipeline no longer supports vsone'
        if qreq_.prog_hook is not None:
            qreq_.prog_hook.initialize_subhooks(5)

        # qreq_.lazy_load(verbose=(verbose and ut.NOT_QUIET))
        qreq_.lazy_preload(verbose=(verbose and ut.NOT_QUIET))
        impossible_daids_list, Kpad_list = build_impossible_daids_list(qreq_)

        # Nearest neighbors (nns_list)
        # a nns object is a tuple(ndarray, ndarray) - (qfx2_dx, qfx2_dist)
        # * query descriptors assigned to database descriptors
        # * FLANN used here
        nns_list = nearest_neighbors(
            qreq_, Kpad_list, impossible_daids_list, verbose=verbose
        )

        # Remove Impossible Votes
        # a nnfilt object is an ndarray qfx2_valid
        # * marks matches to the same image as invalid
        nnvalid0_list = baseline_neighbor_filter(
            qreq_, nns_list, impossible_daids_list, verbose=verbose
        )

        # Nearest neighbors weighting / scoring (filtweights_list)
        # filtweights_list maps qaid to filtweights which is a dict
        # that maps a filter name to that query's weights for that filter
        weight_ret = weight_neighbors(qreq_, nns_list, nnvalid0_list, verbose=verbose)
        filtkey_list, filtweights_list, filtvalids_list, filtnormks_list = weight_ret

        # Nearest neighbors to chip matches (cm_list)
        # * Initial scoring occurs
        # * vsone un-swapping occurs here
        cm_list_FILT = build_chipmatches(
            qreq_,
            nns_list,
            nnvalid0_list,
            filtkey_list,
            filtweights_list,
            filtvalids_list,
            filtnormks_list,
            verbose=verbose,
        )
    else:
        logger.info('invalid pipeline root %r' % (qreq_.qparams.pipeline_root))

    # Spatial verification (cm_list) (TODO: cython)
    # * prunes chip results and feature matches
    # TODO: allow for reweighting of feature matches to happen.
    cm_list_SVER = spatial_verification(qreq_, cm_list_FILT, verbose=verbose)
    if cm_list_FILT[0].filtnorm_aids is not None:
        pass
        # assert cm_list_SVER[0].filtnorm_aids is not None

    cm_list = cm_list_SVER
    # Final Scoring
    score_method = qreq_.qparams.score_method
    scoring.score_chipmatch_list(qreq_, cm_list, score_method)

    if VERB_PIPELINE:
        logger.info('[hs] L___ FINISHED HOTSPOTTER PIPELINE ___')

    return cm_list


# ============================
# 0) Nearest Neighbors
# ============================


def build_impossible_daids_list(qreq_, verbose=VERB_PIPELINE):
    r"""
    Args:
        qreq_ (QueryRequest):  query request object with hyper-parameters

    CommandLine:
        python -m wbia.algo.hots.pipeline --test-build_impossible_daids_list

    Example:
        >>> # ENABLE_DOCTEST
        >>> from wbia.algo.hots.pipeline import *  # NOQA
        >>> import wbia
        >>> qreq_ = wbia.testdata_qreq_(
        >>>     defaultdb='testdb1',
        >>>     a='default:species=zebra_plains,qhackerrors=True',
        >>>     p='default:use_k_padding=True,can_match_sameimg=False,can_match_samename=False')
        >>> impossible_daids_list, Kpad_list = build_impossible_daids_list(qreq_)
        >>> impossible_daids_list = [x.tolist() for x in impossible_daids_list]
        >>> vals = ut.dict_subset(locals(), ['impossible_daids_list', 'Kpad_list'])
        >>> result = ut.repr2(vals, nl=1, explicit=True, nobr=True, strvals=True)
        >>> print(result)
        >>> assert np.all(qreq_.qaids == [1, 4, 5, 6])
        >>> assert np.all(qreq_.daids == [1, 2, 3, 4, 5, 6])
        ...
        impossible_daids_list=[[1], [4], [5, 6], [5, 6]],
        Kpad_list=[1, 1, 2, 2],
    """
    if verbose:
        logger.info('[hs] Step 0) Build impossible matches')

    can_match_sameimg = qreq_.qparams.can_match_sameimg
    can_match_samename = qreq_.qparams.can_match_samename
    use_k_padding = qreq_.qparams.use_k_padding
    can_match_self = False
    internal_qaids = qreq_.get_internal_qaids()
    internal_daids = qreq_.get_internal_daids()
    internal_data_nids = qreq_.get_qreq_annot_nids(internal_daids)

    _impossible_daid_lists = []
    if not can_match_self:
        if can_match_sameimg and can_match_samename:
            # we can skip this if sameimg or samename is specified.
            # it will cover this case for us
            _impossible_daid_lists.append([[qaid] for qaid in internal_qaids])
    if not can_match_sameimg:
        # slow way of getting contact_aids (now incorporates faster way)
        contact_aids_list = qreq_.ibs.get_annot_contact_aids(
            internal_qaids, daid_list=internal_daids
        )
        _impossible_daid_lists.append(contact_aids_list)
        EXTEND_TO_OTHER_CONTACT_GT = False
        # TODO: flag overlapping keypoints with another annot as likely to
        # cause photobombs.
        # Also cannot match any aids with a name of an annotation in this image
        if EXTEND_TO_OTHER_CONTACT_GT:
            # TODO: need a test set that can accomidate testing this case
            # testdb1 might cut it if we spruced it up
            nonself_contact_aids = [
                np.setdiff1d(aids, qaid)
                for aids, qaid in zip(contact_aids_list, internal_qaids)
            ]
            nonself_contact_nids = qreq_.ibs.unflat_map(
                qreq_.get_qreq_annot_nids, nonself_contact_aids
            )
            contact_aids_gt_list = [
                internal_daids.compress(vt.get_covered_mask(internal_data_nids, nids))
                for nids in nonself_contact_nids
            ]
            _impossible_daid_lists.append(contact_aids_gt_list)

    if not can_match_samename:
        internal_data_nids = qreq_.get_qreq_annot_nids(internal_daids)
        internal_query_nids = qreq_.get_qreq_annot_nids(internal_qaids)
        gt_aids = [
            internal_daids.compress(internal_data_nids == nid)
            for nid in internal_query_nids
        ]
        _impossible_daid_lists.append(gt_aids)
    # TODO: add explicit not a match case in here
    _impossible_daids_list = list(map(ut.flatten, zip(*_impossible_daid_lists)))
    impossible_daids_list = [
        np.unique(impossible_daids) for impossible_daids in _impossible_daids_list
    ]

    # TODO: we need to pad K for each bad annotation
    if use_k_padding:
        Kpad_list = list(map(len, impossible_daids_list))
    else:
        # always at least pad K for self queries
        Kpad_list = [1 if qaid in internal_daids else 0 for qaid in internal_qaids]
    return impossible_daids_list, Kpad_list


# ============================
# 1) Nearest Neighbors
# ============================


@profile
def nearest_neighbor_cacheid2(qreq_, Kpad_list):
    r"""
    Returns a hacky cacheid for neighbor configs.
    DEPRICATE: This will be replaced by dtool caching

    Args:
        qreq_ (QueryRequest):  query request object with hyper-parameters
        Kpad_list (list):

    Returns:
        tuple: (nn_mid_cacheid_list, nn_cachedir)

    CommandLine:
        python -m wbia.algo.hots.pipeline --exec-nearest_neighbor_cacheid2
        python -m wbia.algo.hots.pipeline --exec-nearest_neighbor_cacheid2 --superstrict

    Example:
        >>> # ENABLE_DOCTEST
        >>> from wbia.algo.hots.pipeline import *  # NOQA
        >>> import wbia
        >>> verbose = True
        >>> cfgdict = dict(K=4, Knorm=1, checks=800, use_k_padding=False)
        >>> # test 1
        >>> p = 'default' + ut.get_cfg_lbl(cfgdict)
        >>> qreq_ = wbia.testdata_qreq_(
        >>>     defaultdb='testdb1', p=[p], qaid_override=[1, 2],
        >>>     daid_override=[1, 2, 3, 4, 5])
        >>> locals_ = plh.testrun_pipeline_upto(qreq_, 'nearest_neighbors')
        >>> Kpad_list, = ut.dict_take(locals_, ['Kpad_list'])
        >>> tup = nearest_neighbor_cacheid2(qreq_, Kpad_list)
        >>> (nn_cachedir, nn_mid_cacheid_list) = tup
        >>> result1 = 'nn_mid_cacheid_list1 = ' + ut.repr2(nn_mid_cacheid_list, nl=1)
        >>> # test 2
        >>> cfgdict2 = dict(K=2, Knorm=3, use_k_padding=True)
        >>> p2 = 'default' + ut.get_cfg_lbl(cfgdict)
        >>> ibs = qreq_.ibs
        >>> qreq_ = wbia.testdata_qreq_(defaultdb='testdb1', p=[p2], qaid_override=[1, 2], daid_override=[1, 2, 3, 4, 5])
        >>> locals_ = plh.testrun_pipeline_upto(qreq_, 'nearest_neighbors')
        >>> Kpad_list, = ut.dict_take(locals_, ['Kpad_list'])
        >>> tup = nearest_neighbor_cacheid2(qreq_, Kpad_list)
        >>> (nn_cachedir, nn_mid_cacheid_list) = tup
        >>> result2 = 'nn_mid_cacheid_list2 = ' + ut.repr2(nn_mid_cacheid_list, nl=1)
        >>> result = result1 + '\n' + result2
        >>> print(result)
        nn_mid_cacheid_list1 = [
            'nnobj_8687dcb6-1f1f-fdd3-8b72-8f36f9f41905_DVUUIDS((5)oavtblnlrtocnrpm)_NN(single,cks800)_Chip(sz700,maxwh)_Feat(hesaff+sift)_FLANN(8_kdtrees)_truek6',
            'nnobj_a2aef668-20c1-1897-d8f3-09a47a73f26a_DVUUIDS((5)oavtblnlrtocnrpm)_NN(single,cks800)_Chip(sz700,maxwh)_Feat(hesaff+sift)_FLANN(8_kdtrees)_truek6',
        ]
        nn_mid_cacheid_list2 = [
            'nnobj_8687dcb6-1f1f-fdd3-8b72-8f36f9f41905_DVUUIDS((5)oavtblnlrtocnrpm)_NN(single,cks800)_Chip(sz700,maxwh)_Feat(hesaff+sift)_FLANN(8_kdtrees)_truek6',
            'nnobj_a2aef668-20c1-1897-d8f3-09a47a73f26a_DVUUIDS((5)oavtblnlrtocnrpm)_NN(single,cks800)_Chip(sz700,maxwh)_Feat(hesaff+sift)_FLANN(8_kdtrees)_truek6',
        ]
    """
    from wbia.algo import Config

    chip_cfgstr = qreq_.qparams.chip_cfgstr
    feat_cfgstr = qreq_.qparams.feat_cfgstr
    flann_cfgstr = qreq_.qparams.flann_cfgstr
    requery = qreq_.qparams.requery
    # assert requery is False, 'can not be on yet'

    internal_daids = qreq_.get_internal_daids()
    internal_qaids = qreq_.get_internal_qaids()
    if requery:
        assert qreq_.qparams.vsmany
        data_hashid = qreq_.get_data_hashid()
    else:
        data_hashid = qreq_.ibs.get_annot_hashid_visual_uuid(internal_daids, prefix='D')
    if requery:
        query_hashid_list = qreq_.get_qreq_pcc_uuids(internal_qaids)
    else:
        # TODO: get attribute from qreq_, not wbia
        query_hashid_list = qreq_.get_qreq_annot_visual_uuids(internal_qaids)

    HACK_KCFG = True
    if HACK_KCFG:
        # hack config so we consolidate different k values
        # (ie, K=2,Knorm=1 == K=1,Knorm=2)
        nn_cfgstr = Config.NNConfig(**qreq_.qparams).get_cfgstr(
            ignore_keys={'K', 'Knorm', 'use_k_padding'}
        )
    else:
        nn_cfgstr = qreq_.qparams.nn_cfgstr

    aug_cfgstr = 'aug_quryside' if qreq_.qparams.query_rotation_heuristic else ''
    nn_mid_cacheid = ''.join(
        [data_hashid, nn_cfgstr, chip_cfgstr, feat_cfgstr, flann_cfgstr, aug_cfgstr]
    )
    logger.info('nn_mid_cacheid = %r' % (nn_mid_cacheid,))

    if HACK_KCFG:
        kbase = qreq_.qparams.K + int(qreq_.qparams.Knorm)
        nn_mid_cacheid_list = [
            'nnobj_' + str(query_hashid) + nn_mid_cacheid + '_truek' + str(kbase + Kpad)
            for query_hashid, Kpad in zip(query_hashid_list, Kpad_list)
        ]
    else:
        nn_mid_cacheid_list = [
            'nnobj_' + str(query_hashid) + nn_mid_cacheid + '_' + str(Kpad)
            for query_hashid, Kpad in zip(query_hashid_list, Kpad_list)
        ]

    nn_cachedir = qreq_.ibs.get_neighbor_cachedir()
    # ut.unixjoin(qreq_.ibs.get_cachedir(), 'neighborcache2')
    ut.ensuredir(nn_cachedir)
    if ut.VERBOSE:
        logger.info('nn_mid_cacheid = %r' % (nn_mid_cacheid,))
        pass
    return nn_cachedir, nn_mid_cacheid_list


@profile
def cachemiss_nn_compute_fn(
    flags_list, qreq_, Kpad_list, impossible_daids_list, K, Knorm, requery, verbose
):
    """
    Logic for computing neighbors if there is a cache miss

    >>> flags_list = [True] * len(Kpad_list)
    >>> flags_list = [True, False, True]
    """
    # Cant do this here because of get_nn_aids. bleh
    # Could make this slightly more efficient
    # qreq_.load_indexer(verbose=verbose)

    # internal_qaids = qreq_.get_internal_qaids()
    # internal_qaids = internal_qaids.compress(flags_list)

    # Get only the data that needs to be computed
    internal_qannots = qreq_.internal_qannots
    internal_qannots = internal_qannots.compress(flags_list)

    Kpad_list = ut.compress(Kpad_list, flags_list)
    # do computation
    if not requery:
        num_neighbors_list = [K + Kpad + Knorm for Kpad in Kpad_list]
    else:
        num_neighbors_list = [K + Knorm] * len(Kpad_list)
        Kpad_list = 2 * np.array(Kpad_list)
    config2_ = qreq_.get_internal_query_config2()
    qvecs_list = internal_qannots.vecs

    qfxs_list = [np.arange(len(qvecs)) for qvecs in qvecs_list]

    if config2_.minscale_thresh is not None or config2_.maxscale_thresh is not None:
        min_ = -np.inf if config2_.minscale_thresh is None else config2_.minscale_thresh
        max_ = np.inf if config2_.maxscale_thresh is None else config2_.maxscale_thresh
        # qkpts_list = qreq_.ibs.get_annot_kpts(internal_qaids, config2_=config2_)
        qkpts_list = internal_qannots.kpts
        qkpts_list = vt.ziptake(qkpts_list, qfxs_list, axis=0)
        # kpts_list = vt.ziptake(kpts_list, fxs_list, axis=0)  # not needed for first filter
        scales_list = [vt.get_scales(kpts) for kpts in qkpts_list]
        # Remove data under the threshold
        flags_list1 = [
            np.logical_and(scales >= min_, scales <= max_) for scales in scales_list
        ]
        qvecs_list = vt.zipcompress(qvecs_list, flags_list1, axis=0)
        qfxs_list = vt.zipcompress(qfxs_list, flags_list1, axis=0)

    if config2_.fgw_thresh is not None:
        # qfgw_list = qreq_.ibs.get_annot_fgweights(
        #    internal_qaids, config2_=config2_)
        qfgw_list = internal_qannots.fgweights
        qfgw_list = vt.ziptake(qfgw_list, qfxs_list, axis=0)
        fgw_thresh = config2_.fgw_thresh
        flags_list2 = [fgws >= fgw_thresh for fgws in qfgw_list]
        qfxs_list = vt.zipcompress(qfxs_list, flags_list2, axis=0)
        qvecs_list = vt.zipcompress(qvecs_list, flags_list2, axis=0)

    if verbose:
        if len(qvecs_list) == 1:
            logger.info('[hs] depth(qvecs_list) = %r' % (ut.depth_profile(qvecs_list),))
    # Mark progress ane execute nearest indexer nearest neighbor code
    prog_hook = None if qreq_.prog_hook is None else qreq_.prog_hook.next_subhook()
    if requery:
        # assert False, (
        #     'need to implement part where matches with the same name are not considered'
        # )
        qvec_iter = ut.ProgressIter(qvecs_list, lbl=NN_LBL, prog_hook=prog_hook, **PROGKW)

        """
        # Maybe some query vector stacking would help here
        qvecs_stack = np.vstack(qvecs_list)

        # Nope, really doesn't help that much

        # For 100 annotations
        %timeit np.vstack(qvecs_list)
        # 100 loops, best of 3: 5.15 ms per loop
        %timeit qreq_.indexer.knn(qvecs_stack, K)
        # 1 loop, best of 3: 18.1 s per loop
        %timeit [qreq_.indexer.knn(qfx2_vec, K) for qfx2_vec in qvecs_list]
        # 1 loop, best of 3: 19.4 s per loop
        """

        idx_dist_list = [
            qreq_.indexer.requery_knn(qfx2_vec, K, pad, impossible_daids)
            for qfx2_vec, K, pad, impossible_daids in zip(
                qvec_iter, num_neighbors_list, Kpad_list, impossible_daids_list
            )
        ]
    else:
        qvec_iter = ut.ProgressIter(qvecs_list, lbl=NN_LBL, prog_hook=prog_hook, **PROGKW)
        idx_dist_list = [
            qreq_.indexer.knn(qfx2_vec, num_neighbors)
            for qfx2_vec, num_neighbors in zip(qvec_iter, num_neighbors_list)
        ]

    # Move into new object structure
    nns_list = [
        Neighbors(qaid, idxs, dists, qfxs)
        for qaid, qfxs, (idxs, dists) in zip(
            internal_qannots.aid, qfxs_list, idx_dist_list
        )
    ]
    return nns_list


@profile
def nearest_neighbors(
    qreq_, Kpad_list, impossible_daids_list=None, verbose=VERB_PIPELINE
):
    """
    Plain Nearest Neighbors
    Tries to load nearest neighbors from a cache instead of recomputing them.

    CommandLine:
        python -m wbia.algo.hots.pipeline --test-nearest_neighbors
        python -m wbia.algo.hots.pipeline --test-nearest_neighbors --db PZ_MTEST --qaids=1:100
        utprof.py -m wbia.algo.hots.pipeline --test-nearest_neighbors --db PZ_MTEST --qaids=1:100

    Example:
        >>> # ENABLE_DOCTEST
        >>> from wbia.algo.hots.pipeline import *  # NOQA
        >>> import wbia
        >>> verbose = True
        >>> qreq_ = wbia.testdata_qreq_(defaultdb='testdb1', qaid_override=[1])
        >>> locals_ = plh.testrun_pipeline_upto(qreq_, 'nearest_neighbors')
        >>> Kpad_list, impossible_daids_list = ut.dict_take(
        >>>     locals_, ['Kpad_list', 'impossible_daids_list'])
        >>> nns_list = nearest_neighbors(qreq_, Kpad_list, impossible_daids_list,
        >>>                              verbose=verbose)
        >>> qaid = qreq_.internal_qaids[0]
        >>> nn = nns_list[0]
        >>> (qfx2_idx, qfx2_dist) = nn
        >>> num_neighbors = Kpad_list[0] + qreq_.qparams.K + qreq_.qparams.Knorm
        >>> # Assert nns tuple is valid
        >>> ut.assert_eq(qfx2_idx.shape, qfx2_dist.shape)
        >>> ut.assert_eq(qfx2_idx.shape[1], num_neighbors)
        >>> ut.assert_inbounds(qfx2_idx.shape[0], 1000, 3000)

    Example:
        >>> # ENABLE_DOCTEST
        >>> from wbia.algo.hots.pipeline import *  # NOQA
        >>> import wbia
        >>> verbose = True
        >>> qreq_ = wbia.testdata_qreq_(defaultdb='testdb1', qaid_override=[1])
        >>> locals_ = plh.testrun_pipeline_upto(qreq_, 'nearest_neighbors')
        >>> Kpad_list, impossible_daids_list = ut.dict_take(
        >>>     locals_, ['Kpad_list', 'impossible_daids_list'])
        >>> nns_list = nearest_neighbors(qreq_, Kpad_list, impossible_daids_list,
        >>>                              verbose=verbose)
        >>> qaid = qreq_.internal_qaids[0]
        >>> nn = nns_list[0]
        >>> (qfx2_idx, qfx2_dist) = nn
        >>> num_neighbors = Kpad_list[0] + qreq_.qparams.K + qreq_.qparams.Knorm
        >>> # Assert nns tuple is valid
        >>> ut.assert_eq(qfx2_idx.shape, qfx2_dist.shape)
        >>> ut.assert_eq(qfx2_idx.shape[1], num_neighbors)
        >>> ut.assert_inbounds(qfx2_idx.shape[0], 1000, 3000)

    Example:
        >>> # ENABLE_DOCTEST
        >>> from wbia.algo.hots.pipeline import *  # NOQA
        >>> import wbia
        >>> verbose = True
        >>> custom_nid_lookup = {a: a for a in range(14)}
        >>> qreq1_ = wbia.testdata_qreq_(
        >>>     defaultdb='testdb1', t=['default:K=2,requery=True,can_match_samename=False'],
        >>>     daid_override=[2, 3, 4, 5, 6, 7, 8],
        >>>     qaid_override=[2, 5, 1], custom_nid_lookup=custom_nid_lookup)
        >>> locals_ = plh.testrun_pipeline_upto(qreq1_, 'nearest_neighbors')
        >>> Kpad_list, impossible_daids_list = ut.dict_take(
        >>>    locals_, ['Kpad_list', 'impossible_daids_list'])
        >>> nns_list1 = nearest_neighbors(qreq1_, Kpad_list, impossible_daids_list,
        >>>                               verbose=verbose)
        >>> nn1 = nns_list1[0]
        >>> nnvalid0_list1 = baseline_neighbor_filter(qreq1_, nns_list1,
        >>>                                           impossible_daids_list)
        >>> assert np.all(nnvalid0_list1[0]), (
        >>>  'requery should never produce impossible results')
        >>> # Compare versus not using requery
        >>> qreq2_ = wbia.testdata_qreq_(
        >>>     defaultdb='testdb1', t=['default:K=2,requery=False'],
        >>>     daid_override=[1, 2, 3, 4, 5, 6, 7, 8],
        >>>     qaid_override=[2, 5, 1])
        >>> locals_ = plh.testrun_pipeline_upto(qreq2_, 'nearest_neighbors')
        >>> Kpad_list, impossible_daids_list = ut.dict_take(
        >>>    locals_, ['Kpad_list', 'impossible_daids_list'])
        >>> nns_list2 = nearest_neighbors(qreq2_, Kpad_list, impossible_daids_list,
        >>>                               verbose=verbose)
        >>> nn2 = nns_list2[0]
        >>> nn1.neighb_dists
        >>> nn2.neighb_dists

    Example:
        >>> # ENABLE_DOCTEST
        >>> from wbia.algo.hots.pipeline import *  # NOQA
        >>> import wbia
        >>> verbose = True
        >>> qreq1_ = wbia.testdata_qreq_(
        >>>     defaultdb='testdb1', t=['default:K=5,requery=True,can_match_samename=False'],
        >>>     daid_override=[2, 3, 4, 5, 6, 7, 8],
        >>>     qaid_override=[2, 5, 1])
        >>> locals_ = plh.testrun_pipeline_upto(qreq1_, 'nearest_neighbors')
        >>> Kpad_list, impossible_daids_list = ut.dict_take(
        >>>    locals_, ['Kpad_list', 'impossible_daids_list'])
        >>> nns_list1 = nearest_neighbors(qreq1_, Kpad_list, impossible_daids_list,
        >>>                               verbose=verbose)
        >>> nn1 = nns_list1[0]
        >>> nnvalid0_list1 = baseline_neighbor_filter(qreq1_, nns_list1,
        >>>                                           impossible_daids_list)
        >>> assert np.all(nnvalid0_list1[0]), 'should always be valid'
    """
    K = qreq_.qparams.K
    Knorm = qreq_.qparams.Knorm
    requery = qreq_.qparams.requery
    # checks = qreq_.qparams.checks
    # Get both match neighbors (including padding) and normalizing neighbors
    if verbose:
        logger.info(
            '[hs] Step 1) Assign nearest neighbors: %s' % (qreq_.qparams.nn_cfgstr,)
        )

    prog_hook = None if qreq_.prog_hook is None else qreq_.prog_hook.next_subhook()
    qreq_.load_indexer(verbose=verbose, prog_hook=prog_hook)
    # For each internal query annotation
    # Find the nearest neighbors of each descriptor vector
    # USE_NN_MID_CACHE = ut.is_developer()

    use_cache = USE_NN_MID_CACHE
    if use_cache:
        nn_cachedir, nn_mid_cacheid_list = nearest_neighbor_cacheid2(qreq_, Kpad_list)
    else:
        # hacks
        nn_mid_cacheid_list = [None] * len(qreq_.get_internal_qaids())
        nn_cachedir = None

    nns_list = ut.tryload_cache_list_with_compute(
        use_cache,
        nn_cachedir,
        'neighbs4',
        nn_mid_cacheid_list,
        cachemiss_nn_compute_fn,
        qreq_,
        Kpad_list,
        impossible_daids_list,
        K,
        Knorm,
        requery,
        verbose,
    )
    return nns_list


# ============================
# 2) Remove Impossible Weights
# ============================


@profile
def baseline_neighbor_filter(
    qreq_, nns_list, impossible_daids_list, verbose=VERB_PIPELINE
):
    """
    Removes matches to self, the same image, or the same name.

    CommandLine:
        python -m wbia.algo.hots.pipeline --test-baseline_neighbor_filter

    Example:
        >>> # ENABLE_DOCTEST
        >>> from wbia.algo.hots.pipeline import *   # NOQA
        >>> qreq_, args = plh.testdata_pre(
        >>>     'baseline_neighbor_filter', defaultdb='testdb1',
        >>>     qaid_override=[1, 2, 3, 4],
        >>>     daid_override=list(range(1, 11)),
        >>>     p=['default:QRH=False,requery=False,can_match_samename=False'],
        >>>     verbose=True)
        >>> nns_list, impossible_daids_list = args
        >>> nnvalid0_list = baseline_neighbor_filter(qreq_, nns_list,
        >>>                                          impossible_daids_list)
        >>> ut.assert_eq(len(nnvalid0_list), len(qreq_.qaids))
        >>> assert not np.any(nnvalid0_list[0][:, 0]), (
        ...    'first col should be all invalid because of self match')
        >>> assert not np.all(nnvalid0_list[0][:, 1]), (
        ...    'second col should have some good matches')
        >>> ut.assert_inbounds(nnvalid0_list[0].sum(), 1000, 10000)
    """
    if verbose:
        logger.info('[hs] Step 2) Baseline neighbor filter')
    Knorm = qreq_.qparams.Knorm
    # Find which annotations each query matched against
    neighb_aids_iter = (
        qreq_.indexer.get_nn_aids(nn.neighb_idxs.T[0 : nn.neighb_idxs.shape[1] - Knorm].T)
        for nn in nns_list
    )
    filter_iter_ = zip(neighb_aids_iter, impossible_daids_list)
    prog_hook = None if qreq_.prog_hook is None else qreq_.prog_hook.next_subhook()
    filter_iter = ut.ProgressIter(
        filter_iter_,
        length=len(nns_list),
        enabled=False,
        lbl=FILT_LBL,
        prog_hook=prog_hook,
        **PROGKW,
    )
    # Check to be sure that none of the matched annotations are in the impossible set
    nnvalid0_list = [
        vt.get_uncovered_mask(neighb_aids, impossible_daids)
        for neighb_aids, impossible_daids in filter_iter
    ]
    return nnvalid0_list


# ============================
# 3) Nearest Neighbor weights
# ============================


@profile
def weight_neighbors(qreq_, nns_list, nnvalid0_list, verbose=VERB_PIPELINE):
    """
    pipeline step 3 -
    assigns weights to feature matches based on the active filter list

    CommandLine:
        python -m wbia.algo.hots.pipeline --test-weight_neighbors
        python -m wbia.algo.hots.pipeline --test-weight_neighbors:0 --verbose --verbtd --ainfo --nocache --veryverbose
        python -m wbia.algo.hots.pipeline --test-weight_neighbors:0 --show
        python -m wbia.algo.hots.pipeline --test-weight_neighbors:1 --show

        python -m wbia.algo.hots.pipeline --test-weight_neighbors:0 --show -t default:lnbnn_normer=lnbnn_fg_0.9__featscore,lnbnn_norm_thresh=.9

    Example:
        >>> # ENABLE_DOCTEST
        >>> from wbia.algo.hots.pipeline import *  # NOQA
        >>> qreq_, args = plh.testdata_pre(
        >>>     'weight_neighbors', defaultdb='testdb1',
        >>>     a=['default:qindex=0:3,dindex=0:5,hackerrors=False'],
        >>>     p=['default:codename=vsmany,bar_l2_on=True,fg_on=False'], verbose=True)
        >>> nns_list, nnvalid0_list = args
        >>> verbose = True
        >>> weight_ret = weight_neighbors(qreq_, nns_list, nnvalid0_list, verbose)
        >>> filtkey_list, filtweights_list, filtvalids_list, filtnormks_list = weight_ret
        >>> import wbia.plottool as pt
        >>> verbose = True
        >>> cm_list = build_chipmatches(
        >>>     qreq_, nns_list, nnvalid0_list, filtkey_list, filtweights_list,
        >>>     filtvalids_list, filtnormks_list, verbose=verbose)
        >>> ut.quit_if_noshow()
        >>> cm = cm_list[0]
        >>> cm.score_name_nsum(qreq_)
        >>> cm.ishow_analysis(qreq_)
        >>> ut.show_if_requested()

    Example:
        >>> # ENABLE_DOCTEST
        >>> from wbia.algo.hots.pipeline import *  # NOQA
        >>> qreq_, args = plh.testdata_pre(
        >>>     'weight_neighbors', defaultdb='testdb1',
        >>>     a=['default:qindex=0:3,dindex=0:5,hackerrors=False'],
        >>>     p=['default:codename=vsmany,bar_l2_on=True,fg_on=False'], verbose=True)
        >>> nns_list, nnvalid0_list = args
        >>> verbose = True
        >>> weight_ret = weight_neighbors(qreq_, nns_list, nnvalid0_list, verbose)
        >>> filtkey_list, filtweights_list, filtvalids_list, filtnormks_list = weight_ret
        >>> nInternAids = len(qreq_.get_internal_qaids())
        >>> nFiltKeys = len(filtkey_list)
        >>> filtweight_depth = ut.depth_profile(filtweights_list)
        >>> filtvalid_depth = ut.depth_profile(filtvalids_list)
        >>> ut.assert_eq(nInternAids, len(filtweights_list))
        >>> ut.assert_eq(nInternAids, len(filtvalids_list))
        >>> ut.assert_eq(ut.get_list_column(filtweight_depth, 0), [nFiltKeys] * nInternAids)
        >>> ut.assert_eq(filtvalid_depth, (nInternAids, nFiltKeys))
        >>> ut.assert_eq(filtvalids_list, [[None, None], [None, None], [None, None]])
        >>> ut.assert_eq(filtkey_list, [hstypes.FiltKeys.LNBNN, hstypes.FiltKeys.BARL2])
        >>> ut.quit_if_noshow()
        >>> import wbia.plottool as pt
        >>> verbose = True
        >>> cm_list = build_chipmatches(
        >>>     qreq_, nns_list, nnvalid0_list, filtkey_list, filtweights_list,
        >>>     filtvalids_list, filtnormks_list, verbose=verbose)
        >>> cm = cm_list[0]
        >>> cm.score_name_nsum(qreq_)
        >>> cm.ishow_analysis(qreq_)
        >>> ut.show_if_requested()
    """
    if verbose:
        logger.info('[hs] Step 3) Weight neighbors: ' + qreq_.qparams.nnweight_cfgstr)
        if len(nns_list) == 1:
            logger.info('[hs] depth(nns_list) ' + str(ut.depth_profile(nns_list)))

    # logger.info(WEIGHT_LBL)
    # intern_qaid_iter = ut.ProgressIter(internal_qaids, lbl=BUILDCM_LBL,
    #                                   **PROGKW)

    # Build weights for each active filter
    filtkey_list = []
    _filtweight_list = []
    _filtvalid_list = []
    _filtnormk_list = []

    config2_ = qreq_.extern_data_config2

    if not config2_.sqrd_dist_on:
        # Take the square root of the squared distances
        for nns in nns_list:
            nns.neighb_dists = np.sqrt(nns.neighb_dists.astype(np.float64))
        # nns_list_ = [(neighb_idx, np.sqrt(neighb_dist.astype(np.float64)))
        #              for neighb_idx, neighb_dist in nns_list]
        # nns_list = nns_list_

    if config2_.lnbnn_on:
        filtname = 'lnbnn'
        lnbnn_weight_list, normk_list = nn_weights.NN_WEIGHT_FUNC_DICT[filtname](
            nns_list, nnvalid0_list, qreq_
        )

        if config2_.lnbnn_normer is not None:
            logger.info('[hs] normalizing feat scores')
            if qreq_.lnbnn_normer is None:
                qreq_.lnbnn_normer = vt.ScoreNormalizer()
                # qreq_.lnbnn_normer.load(cfgstr=config2_.lnbnn_normer)
                qreq_.lnbnn_normer.fuzzyload(partial_cfgstr=config2_.lnbnn_normer)

            lnbnn_weight_list = [
                qreq_.lnbnn_normer.normalize_scores(s.ravel()).reshape(s.shape)
                for s in lnbnn_weight_list
            ]

            # Thresholding like a champ!
            lnbnn_norm_thresh = config2_.lnbnn_norm_thresh
            # lnbnn_weight_list = [
            #     s * [s > lnbnn_norm_thresh] for s in lnbnn_weight_list
            # ]
            lnbnn_isvalid = [s > lnbnn_norm_thresh for s in lnbnn_weight_list]
            # Softmaxing
            # from scipy.special import expit
            # y = expit(x * 6)
            # lnbnn_weight_list = [
            #     vt.logistic_01(s)
            #     for s in lnbnn_weight_list
            # ]
            filtname += '_norm'
            _filtvalid_list.append(lnbnn_isvalid)  # None means all valid
        else:
            _filtvalid_list.append(None)  # None means all valid

        _filtweight_list.append(lnbnn_weight_list)
        _filtnormk_list.append(normk_list)
        filtkey_list.append(filtname)
    if config2_.normonly_on:
        filtname = 'normonly'
        normonly_weight_list, normk_list = nn_weights.NN_WEIGHT_FUNC_DICT[filtname](
            nns_list, nnvalid0_list, qreq_
        )
        _filtweight_list.append(normonly_weight_list)
        _filtvalid_list.append(None)  # None means all valid
        _filtnormk_list.append(normk_list)
        filtkey_list.append(filtname)
    if config2_.bar_l2_on:
        filtname = 'bar_l2'
        bar_l2_weight_list, normk_list = nn_weights.NN_WEIGHT_FUNC_DICT[filtname](
            nns_list, nnvalid0_list, qreq_
        )
        _filtweight_list.append(bar_l2_weight_list)
        _filtvalid_list.append(None)  # None means all valid
        _filtnormk_list.append(None)
        filtkey_list.append(filtname)
    if config2_.ratio_thresh:
        filtname = 'ratio'
        ratio_weight_list, normk_list = nn_weights.NN_WEIGHT_FUNC_DICT[filtname](
            nns_list, nnvalid0_list, qreq_
        )
        ratio_isvalid = [
            neighb_ratio <= qreq_.qparams.ratio_thresh
            for neighb_ratio in ratio_weight_list
        ]
        # HACK TO GET 1 - RATIO AS SCORE
        ratioscore_list = [
            np.subtract(1, neighb_ratio) for neighb_ratio in ratio_weight_list
        ]
        _filtweight_list.append(ratioscore_list)
        _filtvalid_list.append(ratio_isvalid)
        _filtnormk_list.append(normk_list)
        filtkey_list.append(filtname)
    # --simple weighted implm
    if config2_.const_on:
        filtname = 'const'
        constvote_weight_list = nn_weights.NN_WEIGHT_FUNC_DICT[filtname](
            nns_list, nnvalid0_list, qreq_
        )
        _filtweight_list.append(constvote_weight_list)
        _filtvalid_list.append(None)  # None means all valid
        _filtnormk_list.append(None)
        filtkey_list.append(filtname)
    if config2_.fg_on:
        filtname = 'fg'
        fgvote_weight_list = nn_weights.NN_WEIGHT_FUNC_DICT[filtname](
            nns_list, nnvalid0_list, qreq_
        )
        _filtweight_list.append(fgvote_weight_list)
        _filtvalid_list.append(None)  # None means all valid
        _filtnormk_list.append(None)
        filtkey_list.append(filtname)

    # Switch nested list structure from [filt, qaid] to [qaid, filt]
    nInternAids = len(nns_list)
    filtweights_list = [
        ut.get_list_column(_filtweight_list, index) for index in range(nInternAids)
    ]
    filtvalids_list = [
        [None if filtvalid is None else filtvalid[index] for filtvalid in _filtvalid_list]
        for index in range(nInternAids)
    ]

    filtnormks_list = [
        [None if normk is None else normk[index] for normk in _filtnormk_list]
        for index in range(nInternAids)
    ]

    assert len(filtkey_list) > 0, 'no feature correspondece filter keys were specified'

    weight_ret = WeightRet_(
        filtkey_list, filtweights_list, filtvalids_list, filtnormks_list
    )
    return weight_ret


# ============================
# 4) Conversion from featurematches to chipmatches neighb -> aid2
# ============================


# @profile
def build_chipmatches(
    qreq_,
    nns_list,
    nnvalid0_list,
    filtkey_list,
    filtweights_list,
    filtvalids_list,
    filtnormks_list,
    verbose=VERB_PIPELINE,
):
    """
    pipeline step 4 - builds sparse chipmatches

    Takes the dense feature matches from query feature to (what could be any)
    database features and builds sparse matching pairs for each annotation to
    annotation match.

    CommandLine:
        python -m wbia build_chipmatches
        python -m wbia build_chipmatches:0 --show
        python -m wbia build_chipmatches:1 --show
        python -m wbia build_chipmatches:2 --show

    Example0:
        >>> # ENABLE_DOCTEST
        >>> from wbia.algo.hots.pipeline import *  # NOQA
        >>> qreq_, args = plh.testdata_pre(
        >>>     'build_chipmatches', p=['default:codename=vsmany'])
        >>> (nns_list, nnvalid0_list, filtkey_list, filtweights_list,
        >>> filtvalids_list, filtnormks_list) = args
        >>> verbose = True
        >>> cm_list = build_chipmatches(qreq_, *args, verbose=verbose)
        >>> # verify results
        >>> [cm.assert_self(qreq_) for cm in cm_list]
        >>> cm = cm_list[0]
        >>> fm = cm.fm_list[cm.daid2_idx[2]]
        >>> num_matches = len(fm)
        >>> print('vsmany num_matches = %r' % num_matches)
        >>> ut.assert_inbounds(num_matches, 500, 2000, 'vsmany nmatches out of bounds')
        >>> ut.quit_if_noshow()
        >>> cm.score_annot_csum(qreq_)
        >>> cm_list[0].ishow_single_annotmatch(qreq_)
        >>> ut.show_if_requested()

    Example1:
        >>> # ENABLE_DOCTEST
        >>> from wbia.algo.hots.pipeline import *  # NOQA
        >>> # Test to make sure filtering by feature weights works
        >>> qreq_, args = plh.testdata_pre(
        >>>     'build_chipmatches',
        >>>     p=['default:codename=vsmany,fgw_thresh=.9'])
        >>> (nns_list, nnvalid0_list, filtkey_list, filtweights_list,
        >>>  filtvalids_list, filtnormks_list) = args
        >>> verbose = True
        >>> cm_list = build_chipmatches(qreq_, *args, verbose=verbose)
        >>> # verify results
        >>> [cm.assert_self(qreq_) for cm in cm_list]
        >>> cm = cm_list[0]
        >>> fm = cm.fm_list[cm.daid2_idx[2]]
        >>> num_matches = len(fm)
        >>> print('num_matches = %r' % num_matches)
        >>> ut.assert_inbounds(num_matches, 100, 410, 'vsmany nmatches out of bounds')
        >>> ut.quit_if_noshow()
        >>> cm.score_annot_csum(qreq_)
        >>> cm_list[0].ishow_single_annotmatch(qreq_)
        >>> ut.show_if_requested()

    Example2:
        >>> # ENABLE_DOCTEST
        >>> from wbia.algo.hots.pipeline import *  # NOQA
        >>> qreq_, args = plh.testdata_pre(
        >>>     'build_chipmatches', p=['default:requery=True'], a='default')
        >>> (nns_list, nnvalid0_list, filtkey_list, filtweights_list,
        >>> filtvalids_list, filtnormks_list) = args
        >>> verbose = True
        >>> cm_list = build_chipmatches(qreq_, *args, verbose=verbose)
        >>> # verify results
        >>> [cm.assert_self(qreq_) for cm in cm_list]
        >>> scoring.score_chipmatch_list(qreq_, cm_list, 'csum')
        >>> cm = cm_list[0]
        >>> for cm in cm_list:
        >>>     # should be positive for LNBNN
        >>>     assert np.all(cm.score_list[np.isfinite(cm.score_list)] >= 0)
    """
    assert not qreq_.qparams.vsone, 'can no longer do vsone in pipeline'
    Knorm = qreq_.qparams.Knorm
    if verbose:
        pipeline_root = qreq_.qparams.pipeline_root
        logger.info('[hs] Step 4) Building chipmatches %s' % (pipeline_root,))

    # Iterate over INTERNAL query annotation ids
    prog_hook = None if qreq_.prog_hook is None else qreq_.prog_hook.next_subhook()
    nns_iter = ut.ProgressIter(
        nns_list, lbl=BUILDCM_LBL, enabled=False, prog_hook=prog_hook, **PROGKW
    )

    cm_list = [
        get_sparse_matchinfo_nonagg(
            qreq_,
            nns,
            neighb_valid0,
            neighb_score_list,
            neighb_valid_list,
            neighb_normk_list,
            Knorm,
            fsv_col_lbls=filtkey_list,
        )
        for nns, neighb_valid0, neighb_score_list, neighb_valid_list, neighb_normk_list in zip(
            nns_iter, nnvalid0_list, filtweights_list, filtvalids_list, filtnormks_list
        )
    ]
    return cm_list


# @profile
def get_sparse_matchinfo_nonagg(
    qreq_,
    nns,
    neighb_valid0,
    neighb_score_list,
    neighb_valid_list,
    neighb_normk_list,
    Knorm,
    fsv_col_lbls,
):
    """
    builds sparse iterator that generates feature match pairs, scores, and ranks

    Returns:
        ValidMatchTup_ : vmt a tuple of corresponding lists. Each item in the
            list corresponds to a daid, dfx, scorevec, rank, norm_aid, norm_fx...

    CommandLine:
        python -m wbia.algo.hots.pipeline --test-get_sparse_matchinfo_nonagg --show
        python -m wbia.algo.hots.pipeline --test-get_sparse_matchinfo_nonagg:1 --show

        utprof.py -m wbia.algo.hots.pipeline --test-get_sparse_matchinfo_nonagg

    Example0:
        >>> # ENABLE_DOCTEST
        >>> from wbia.algo.hots.pipeline import *  # NOQA
        >>> verbose = True
        >>> qreq_, qaid, daid, args = plh.testdata_sparse_matchinfo_nonagg(
        >>>     defaultdb='PZ_MTEST', p=['default:Knorm=3,normalizer_rule=name,const_on=True,ratio_thresh=.2,sqrd_dist_on=True'])
        >>> nns, neighb_valid0, neighb_score_list, neighb_valid_list, neighb_normk_list, Knorm, fsv_col_lbls = args
        >>> cm = get_sparse_matchinfo_nonagg(qreq_, *args)
        >>> qannot = qreq_.ibs.annots([qaid], config=qreq_.qparams)
        >>> dannot = qreq_.ibs.annots(cm.daid_list, config=qreq_.qparams)
        >>> cm.assert_self(verbose=False)
        >>> ut.quit_if_noshow()
        >>> cm.score_annot_csum(qreq_)
        >>> cm.show_single_annotmatch(qreq_)
        >>> ut.show_if_requested()
    """
    # Unpack neighbor ids, indices, filter scores, and flags
    indexer = qreq_.indexer
    neighb_idx = nns.neighb_idxs
    neighb_nnidx = neighb_idx.T[:-Knorm].T
    qfx_list = nns.qfx_list
    K = neighb_nnidx.T.shape[0]
    neighb_daid = indexer.get_nn_aids(neighb_nnidx)
    neighb_dfx = indexer.get_nn_featxs(neighb_nnidx)

    # Determine matches that are valid using all measurements
    neighb_valid_list_ = [neighb_valid0] + ut.filter_Nones(neighb_valid_list)
    neighb_valid_agg = np.logical_and.reduce(neighb_valid_list_)

    # We fill filter each relavant matrix by aggregate validity
    flat_validx = np.flatnonzero(neighb_valid_agg)
    # Infer the valid internal query feature indexes and ranks
    valid_x = np.floor_divide(flat_validx, K, dtype=hstypes.INDEX_TYPE)
    valid_qfx = qfx_list.take(valid_x)
    valid_rank = np.mod(flat_validx, K, dtype=hstypes.FK_DTYPE)
    # TODO: valid_qfx, valid_rank = np.unravel_index(flat_validx, (neighb_nnidx.shape[0], K))?
    # Then take the valid indices from internal database
    # annot_rowids, feature indexes, and all scores
    valid_daid = neighb_daid.take(flat_validx, axis=None)
    valid_dfx = neighb_dfx.take(flat_validx, axis=None)
    valid_scorevec = np.concatenate(
        [neighb_score.take(flat_validx)[:, None] for neighb_score in neighb_score_list],
        axis=1,
    )

    # Incorporate Normalizers
    # Normalizers for each weight filter that used a normalizer
    # Determine which feature per annot was used as the normalizer for each filter
    # Each non-None sub list is still in neighb_ format
    num_filts = len(neighb_normk_list)
    K = len(neighb_idx.T) - Knorm
    norm_filtxs = ut.where_not_None(neighb_normk_list)
    num_normed_filts = len(norm_filtxs)
    if num_normed_filts > 0:
        _normks = ut.take(neighb_normk_list, norm_filtxs)
        # Offset index to get flat normalizer positions
        _offset = np.arange(0, neighb_idx.size, neighb_idx.shape[1])
        flat_normxs = [_offset + neighb_normk for neighb_normk in _normks]
        flat_normidxs = [neighb_idx.take(ks) for ks in flat_normxs]
        flat_norm_aids = [indexer.get_nn_aids(idx) for idx in flat_normidxs]
        flat_norm_fxs = [indexer.get_nn_featxs(idx) for idx in flat_normidxs]
        # Take the valid indicies
        _valid_norm_aids = [aids.take(valid_x) for aids in flat_norm_aids]
        _valid_norm_fxs = [fxs.take(valid_x) for fxs in flat_norm_fxs]
    else:
        _valid_norm_aids = []
        _valid_norm_fxs = []
    valid_norm_aids = ut.ungroup([_valid_norm_aids], [norm_filtxs], num_filts - 1)
    valid_norm_fxs = ut.ungroup([_valid_norm_fxs], [norm_filtxs], num_filts - 1)

    # ValidMatchTup_ = namedtuple('vmt', (  # valid_match_tup
    #     'daid', 'qfx', 'dfx', 'scorevec', 'rank', 'norm_aids', 'norm_fxs'))
    # vmt = ValidMatchTup_(valid_daid, valid_qfx, valid_dfx, valid_scorevec,
    #                      valid_rank, valid_norm_aids, valid_norm_fxs)
    # NOTE: CONTIGUOUS ARRAYS MAKE A HUGE DIFFERENCE
    valid_fm = np.concatenate((valid_qfx[:, None], valid_dfx[:, None]), axis=1)
    assert valid_fm.flags.c_contiguous, 'non-contiguous'
    # valid_fm = np.ascontiguousarray(valid_fm)
    daid_list, daid_groupxs = vt.group_indices(valid_daid)

    fm_list = vt.apply_grouping(valid_fm, daid_groupxs)
    fsv_list = vt.apply_grouping(valid_scorevec, daid_groupxs)
    fk_list = vt.apply_grouping(valid_rank, daid_groupxs)

    filtnorm_aids = [
        None  # [None] * len(daid_groupxs)
        if aids is None
        else vt.apply_grouping(aids, daid_groupxs)
        for aids in valid_norm_aids
    ]

    filtnorm_fxs = [
        None  # [None] * len(daid_groupxs)
        if fxs is None
        else vt.apply_grouping(fxs, daid_groupxs)
        for fxs in valid_norm_fxs
    ]

    assert len(filtnorm_aids) == len(fsv_col_lbls), 'bad normer'
    assert len(filtnorm_fxs) == len(fsv_col_lbls), 'bad normer'

    cm = chip_match.ChipMatch(
        nns.qaid,
        daid_list,
        fm_list,
        fsv_list,
        fk_list,
        fsv_col_lbls=fsv_col_lbls,
        filtnorm_aids=filtnorm_aids,
        filtnorm_fxs=filtnorm_fxs,
    )
    return cm


# ============================
# 5) Spatial Verification
# ============================


def spatial_verification(qreq_, cm_list_FILT, verbose=VERB_PIPELINE):
    r"""
    pipeline step 5 - spatially verify feature matches

    Returns:
        list: cm_listSVER - new list of spatially verified chipmatches

    CommandLine:
        python -m wbia.algo.hots.pipeline --test-spatial_verification --show
        python -m wbia.algo.hots.pipeline --test-spatial_verification --show --qaid 1
        python -m wbia.algo.hots.pipeline --test-spatial_verification:0

    Example:
        >>> # ENABLE_DOCTEST
        >>> from wbia.algo.hots.pipeline import *  # NOQA
        >>> ibs, qreq_, cm_list = plh.testdata_pre_sver('PZ_MTEST', qaid_list=[18])
        >>> scoring.score_chipmatch_list(qreq_, cm_list, qreq_.qparams.prescore_method)  # HACK
        >>> cm = cm_list[0]
        >>> top_nids = cm.get_top_nids(6)
        >>> verbose = True
        >>> cm_list_SVER = spatial_verification(qreq_, cm_list)
        >>> # Test Results
        >>> cmSV = cm_list_SVER[0]
        >>> scoring.score_chipmatch_list(qreq_, cm_list_SVER, qreq_.qparams.score_method)  # HACK
        >>> top_nids_SV = cmSV.get_top_nids(6)
        >>> cm.print_csv(sort=True)
        >>> cmSV.print_csv(sort=False)
        >>> gt_daids  = np.intersect1d(cm.get_groundtruth_daids(), cmSV.get_groundtruth_daids())
        >>> fm_list   = cm.get_annot_fm(gt_daids)
        >>> fmSV_list = cmSV.get_annot_fm(gt_daids)
        >>> maplen = lambda list_: np.array(list(map(len, list_)))
        >>> assert len(gt_daids) > 0, 'ground truth did not survive'
        >>> ut.assert_lessthan(maplen(fmSV_list), maplen(fm_list)), 'feature matches were not filtered'
        >>> ut.quit_if_noshow()
        >>> cmSV.show_daids_matches(qreq_, gt_daids)
        >>> import wbia.plottool as pt
        >>> #homog_tup = (refined_inliers, H)
        >>> #aff_tup = (aff_inliers, Aff)
        >>> #pt.draw_sv.show_sv(rchip1, rchip2, kpts1, kpts2, fm, aff_tup=aff_tup, homog_tup=homog_tup, refine_method=refine_method)
        >>> ut.show_if_requested()
    """
    cm_list = cm_list_FILT
    if not qreq_.qparams.sv_on or qreq_.qparams.xy_thresh is None:
        if verbose:
            logger.info('[hs] Step 5) Spatial verification: off')
        return cm_list
    else:
        cm_list_SVER = _spatial_verification(qreq_, cm_list, verbose=verbose)
        return cm_list_SVER


# @profile
def _spatial_verification(qreq_, cm_list, verbose=VERB_PIPELINE):
    """
    make only spatially valid features survive
        >>> from wbia.algo.hots.pipeline import *  # NOQA
    """
    if verbose:
        logger.info('[hs] Step 5) Spatial verification: ' + qreq_.qparams.sv_cfgstr)

    # dbg info (can remove if there is a speed issue)
    score_method = qreq_.qparams.score_method
    prescore_method = qreq_.qparams.prescore_method
    nNameShortList = qreq_.qparams.nNameShortlistSVER
    nAnnotPerName = qreq_.qparams.nAnnotPerNameSVER

    scoring.score_chipmatch_list(qreq_, cm_list, prescore_method)
    cm_shortlist = scoring.make_chipmatch_shortlists(
        qreq_, cm_list, nNameShortList, nAnnotPerName, score_method
    )
    prog_hook = None if qreq_.prog_hook is None else qreq_.prog_hook.next_subhook()
    cm_progiter = ut.ProgressIter(
        cm_shortlist,
        length=len(cm_shortlist),
        prog_hook=prog_hook,
        lbl=SVER_LVL,
        **PROGKW,
    )

    cm_list_SVER = [sver_single_chipmatch(qreq_, cm) for cm in cm_progiter]
    # rescore after verification?
    return cm_list_SVER


# @profile
def sver_single_chipmatch(qreq_, cm, verbose=False):
    r"""
    Spatially verifies a shortlist of a single chipmatch

    TODO: move to chip match?

    loops over a shortlist of results for a specific query annotation

    Args:
        qreq_ (QueryRequest):  query request object with hyper-parameters
        cm (ChipMatch):

    Returns:
        wbia.ChipMatch: cmSV

    CommandLine:
        python -m wbia draw_rank_cmc --db PZ_Master1 --show \
            -t best:refine_method=[homog,affine,cv2-homog,cv2-ransac-homog,cv2-lmeds-homog] \
            -a timectrlhard ---acfginfo --veryverbtd

        python -m wbia draw_rank_cmc --db PZ_Master1 --show \
            -t best:refine_method=[homog,cv2-lmeds-homog],full_homog_checks=[True,False] \
            -a timectrlhard ---acfginfo --veryverbtd

        python -m wbia sver_single_chipmatch --show \
            -t default:full_homog_checks=True -a default --qaid 18

        python -m wbia sver_single_chipmatch --show \
            -t default:refine_method=affine -a default --qaid 18

        python -m wbia sver_single_chipmatch --show \
            -t default:refine_method=cv2-homog -a default --qaid 18

        python -m wbia sver_single_chipmatch --show \
            -t default:refine_method=cv2-homog,full_homog_checks=True -a default --qaid 18

        python -m wbia sver_single_chipmatch --show \
            -t default:refine_method=cv2-homog,full_homog_checks=False -a default --qaid 18

        python -m wbia sver_single_chipmatch --show \
            -t default:refine_method=cv2-lmeds-homog,full_homog_checks=False -a default --qaid 18

        python -m wbia sver_single_chipmatch --show \
            -t default:refine_method=cv2-ransac-homog,full_homog_checks=False -a default --qaid 18

        python -m wbia sver_single_chipmatch --show \
            -t default:full_homog_checks=False -a default --qaid 18

        python -m wbia sver_single_chipmatch --show --qaid=18 --y=0
        python -m wbia sver_single_chipmatch --show --qaid=18 --y=1

    Example:
        >>> # DISABLE_DOCTEST
        >>> # Visualization
        >>> from wbia.algo.hots.pipeline import *  # NOQA
        >>> qreq_, args = plh.testdata_pre('spatial_verification', defaultdb='PZ_MTEST')  #, qaid_list=[18])
        >>> cm_list = args.cm_list_FILT
        >>> ibs = qreq_.ibs
        >>> cm = cm_list[0]
        >>> scoring.score_chipmatch_list(qreq_, cm_list, qreq_.qparams.prescore_method)  # HACK
        >>> #locals_ = ut.exec_func_src(sver_single_chipmatch, key_list=['svtup_list'], sentinal='# <SENTINAL>')
        >>> #svtup_list1, = locals_
        >>> verbose = True
        >>> source = ut.get_func_sourcecode(sver_single_chipmatch, stripdef=True, strip_docstr=True)
        >>> source = ut.replace_between_tags(source, '', '# <SENTINAL>', '# </SENTINAL>')
        >>> globals_ = globals().copy()
        >>> exec(source, globals_)
        >>> svtup_list = globals_['svtup_list']
        >>> gt_daids = cm.get_groundtruth_daids()
        >>> x = ut.get_argval('--y', type_=int, default=0)
        >>> #print('x = %r' % (x,))
        >>> #daid = daids[x % len(daids)]
        >>> notnone_list = ut.not_list(ut.flag_None_items(svtup_list))
        >>> valid_idxs = np.where(notnone_list)
        >>> valid_daids = cm.daid_list[valid_idxs]
        >>> assert len(valid_daids) > 0, 'cannot spatially verify'
        >>> valid_gt_daids = np.intersect1d(gt_daids, valid_daids)
        >>> #assert len(valid_gt_daids) == 0, 'no sver groundtruth'
        >>> daid = valid_gt_daids[x] if len(valid_gt_daids) > 0 else valid_daids[x]
        >>> idx = cm.daid2_idx[daid]
        >>> svtup = svtup_list[idx]
        >>> assert svtup is not None, 'SV TUP IS NONE'
        >>> refined_inliers, refined_errors, H = svtup[0:3]
        >>> aff_inliers, aff_errors, Aff = svtup[3:6]
        >>> homog_tup = (refined_inliers, H)
        >>> aff_tup = (aff_inliers, Aff)
        >>> fm = cm.fm_list[idx]
        >>> aid1 = cm.qaid
        >>> aid2 = daid
        >>> rchip1, = ibs.get_annot_chips([aid1], config2_=qreq_.extern_query_config2)
        >>> kpts1,  = ibs.get_annot_kpts([aid1], config2_=qreq_.extern_query_config2)
        >>> rchip2, = ibs.get_annot_chips([aid2], config2_=qreq_.extern_data_config2)
        >>> kpts2, = ibs.get_annot_kpts([aid2], config2_=qreq_.extern_data_config2)
        >>> import wbia.plottool as pt
        >>> import matplotlib as mpl
        >>> from wbia.scripts.thesis import TMP_RC
        >>> mpl.rcParams.update(TMP_RC)
        >>> show_aff = not ut.get_argflag('--noaff')
        >>> refine_method = qreq_.qparams.refine_method if not ut.get_argflag('--norefinelbl') else ''
        >>> pt.draw_sv.show_sv(rchip1, rchip2, kpts1, kpts2, fm, aff_tup=aff_tup,
        >>>                    homog_tup=homog_tup, show_aff=show_aff,
        >>>                    refine_method=refine_method)
        >>> ut.show_if_requested()
    """
    qaid = cm.qaid
    use_chip_extent = qreq_.qparams.use_chip_extent
    xy_thresh = qreq_.qparams.xy_thresh
    scale_thresh = qreq_.qparams.scale_thresh
    ori_thresh = qreq_.qparams.ori_thresh
    min_nInliers = qreq_.qparams.min_nInliers
    full_homog_checks = qreq_.qparams.full_homog_checks
    refine_method = qreq_.qparams.refine_method
    sver_output_weighting = qreq_.qparams.sver_output_weighting
    # Precompute sver cmtup_old
    kpts1 = qreq_.get_qreq_qannot_kpts(qaid).astype(np.float64)
    kpts2_list = qreq_.get_qreq_dannot_kpts(cm.daid_list)

    if use_chip_extent:
        top_dlen_sqrd_list = qreq_.ibs.get_annot_chip_dlensqrd(
            cm.daid_list, config2_=qreq_.extern_data_config2
        )
    else:
        top_dlen_sqrd_list = compute_matching_dlen_extent(qreq_, cm.fm_list, kpts2_list)
    config2_ = qreq_.extern_query_config2
    if qreq_.qparams.weight_inliers:
        # Weights for inlier scoring
        if config2_.get('fg_on'):
            qweights = qreq_.ibs.get_annot_fgweights(
                [qaid], ensure=True, config2_=config2_
            )[0].astype(np.float64)
        else:
            num = qreq_.ibs.get_annot_num_feats([qaid], config2_=config2_)[0]
            qweights = np.ones(num, np.float64)
        match_weight_list = [qweights.take(fm.T[0]) for fm in cm.fm_list]
    else:
        match_weight_list = [np.ones(len(fm), dtype=np.float64) for fm in cm.fm_list]

    # Make an svtup for every daid in the shortlist
    _iter1 = zip(
        cm.daid_list,
        cm.fm_list,
        cm.fsv_list,
        kpts2_list,
        top_dlen_sqrd_list,
        match_weight_list,
    )
    if verbose:
        _iter1 = ut.ProgIter(
            _iter1, length=len(cm.daid_list), lbl='sver shortlist', freq=1
        )
    svtup_list = []
    for daid, fm, fsv, kpts2, dlen_sqrd2, match_weights in _iter1:
        if len(fm) == 0:
            # skip results without any matches
            sv_tup = None
        else:

            # sver_testdata = dict(
            #     kpts1=kpts1,
            #     kpts2=kpts2,
            #     fm=fm,
            #     xy_thresh=xy_thresh,
            #     scale_thresh=scale_thresh,
            #     ori_thresh=ori_thresh,
            #     dlen_sqrd2=dlen_sqrd2,
            #     min_nInliers=min_nInliers,
            #     match_weights=match_weights,
            #     full_homog_checks=full_homog_checks,
            #     refine_method=refine_method
            # )

            # locals().update(ut.load_data('sver_testdata.pkl'))

            try:
                # Compute homography from chip2 to chip1 returned homography
                # maps image1 space into image2 space image1 is a query chip
                # and image2 is a database chip
                sv_tup = vt.spatially_verify_kpts(
                    kpts1,
                    kpts2,
                    fm,
                    xy_thresh,
                    scale_thresh,
                    ori_thresh,
                    dlen_sqrd2,
                    min_nInliers,
                    match_weights=match_weights,
                    full_homog_checks=full_homog_checks,
                    refine_method=refine_method,
                    returnAff=True,
                )
            except Exception as ex:
                ut.printex(
                    ex,
                    'Unknown error in spatial verification.',
                    keys=[
                        'kpts1',
                        'kpts2',
                        'fm',
                        'xy_thresh',
                        'scale_thresh',
                        'dlen_sqrd2',
                        'min_nInliers',
                    ],
                )
                sv_tup = None
        svtup_list.append(sv_tup)

    # <SENTINAL>

    # New way
    inliers_list = []
    for sv_tup in svtup_list:
        if sv_tup is None:
            inliers_list.append(None)
        else:
            (homog_inliers, homog_errors, H, aff_inliers, aff_errors, Aff) = sv_tup
            inliers_list.append(homog_inliers)

    indicies_list = inliers_list
    cmSV = cm.take_feature_matches(indicies_list, keepscores=False)

    # NOTE: It is not very clear explicitly, but the way H_list and
    # homog_err_weight_list are built will correspond with the daid_list in
    # cmSV returned by cm.take_feature_matches
    svtup_list_ = ut.filter_Nones(svtup_list)
    H_list_SV = ut.get_list_column(svtup_list_, 2)
    cmSV.H_list = H_list_SV

    if sver_output_weighting:
        homog_err_weight_list = []
        xy_thresh_sqrd = dlen_sqrd2 * xy_thresh
        for sv_tup in svtup_list_:
            (homog_inliers, homog_errors) = sv_tup[0:2]
            homog_xy_errors = homog_errors[0].take(homog_inliers, axis=0)
            homog_err_weight = 1.0 - np.sqrt(homog_xy_errors / xy_thresh_sqrd)
            homog_err_weight_list.append(homog_err_weight)
        # Rescore based on homography errors
        filtkey = hstypes.FiltKeys.HOMOGERR
        filtweight_list = homog_err_weight_list
        cmSV.append_featscore_column(filtkey, filtweight_list)
    return cmSV


def compute_matching_dlen_extent(qreq_, fm_list, kpts_list):
    r"""
    helper for spatial verification, computes the squared diagonal length of
    matching chips

    CommandLine:
        python -m wbia.algo.hots.pipeline --test-compute_matching_dlen_extent

    Example:
        >>> # ENABLE_DOCTEST
        >>> from wbia.algo.hots.pipeline import *  # NOQA
        >>> ibs, qreq_, cm_list = plh.testdata_pre_sver('PZ_MTEST')
        >>> verbose = True
        >>> cm = cm_list[0]
        >>> cm.set_cannonical_annot_score(cm.get_num_matches_list())
        >>> cm.sortself()
        >>> fm_list = cm.fm_list
        >>> kpts_list = qreq_.get_qreq_dannot_kpts(cm.daid_list.tolist())
        >>> topx2_dlen_sqrd = compute_matching_dlen_extent(qreq_, fm_list, kpts_list)
        >>> ut.assert_inbounds(np.sqrt(topx2_dlen_sqrd)[0:5], 600, 1500)

    """
    # Use extent of matching keypoints
    # first get matching keypoints
    fx2_list = [fm.T[1] for fm in fm_list]
    kpts2_m_list = vt.ziptake(kpts_list, fx2_list, axis=0)
    # [kpts.take(fx2, axis=0) for (kpts, fx2) in zip(kpts_list, fx2_list)]
    dlen_sqrd_list = [vt.get_kpts_dlen_sqrd(kpts2_m) for kpts2_m in kpts2_m_list]
    return dlen_sqrd_list
